from ..structure import array_schema

OPERATION_STATUS_SCHEMA = {
    'type': 'string',
    'enum': ['WAITING', 'ADDED', 'STARTED', 'SUCCEEDED', 'FAILED', 'UNKNOWN']
}

OPERATION_STATUS_INFO_SCHEMA = {
    "type": "object",
    "properties": {
        "operationId": {
            "type": "string"
        },
        "name": {
            "type": "string"
        },
        "generation": {
            "type": "number"
        },
        "status": OPERATION_STATUS_SCHEMA,
        "added": {
            "type": ["number", "null"]
        },
        "completed": {
            "type": ["number", "null"]
        },
        "errorCode": {
            "type": ["string", "null"]
        },
        "errorMessage": {
            "type": ["string", "null"]
        }
    },
    "required": ["operationId", "name", "generation", "status"],
    "additionalProperties": False
}

REGISTERED_NAME_INFO_SCHEMA = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string"
        },
        "generation": {
            "type": "number"
        },
        "updatingKey": {
            "type": ["string", "null"]
        },
        "nodeUri": {
            "type": "string"
        },
        "created": {
            "type": ["number", "null"]
        },
        "signingKey": {
            "type": ["string", "null"]
        },
        "validFrom": {
            "type": ["number", "null"]
        },
        "digest": {
            "type": ["string", "null"]
        }
    },
    "required": ["name", "generation", "nodeUri"]
}

REGISTERED_NAME_INFO_ARRAY_SCHEMA = array_schema(REGISTERED_NAME_INFO_SCHEMA)

SIGNING_KEY_INFO_SCHEMA = {
    "type": "object",
    "properties": {
        "key": {
            "type": "string"
        },
        "validFrom": {
            "type": "number"
        }
    },
    "required": ["key", "validFrom"],
    "additionalProperties": False
}

SIGNING_KEY_INFO_ARRAY_SCHEMA = array_schema(SIGNING_KEY_INFO_SCHEMA)



import os
from base64 import urlsafe_b64encode
from time import time
from typing import List

from cryptography.hazmat.primitives.asymmetric import ec

from .caller import CarteSource
from .fingerprints import create_carte_fingerprint2
from .node import MoeraNode
from .types import CarteInfo, CarteAttributes, Scope, Timestamp, SCOPE_VALUES
from ..crypto import sign_fingerprint


class MoeraCartesError(Exception):
    """Error obtaining valid cartes."""
    pass


class MoeraCarteSource(CarteSource):
    """Class that gets cartes from the given node, caches them and supplies them for authentication."""
    _node: MoeraNode
    _client_scope: List[Scope]
    _admin_scope: List[Scope]
    _target_node_name: str | None
    _cartes: List[CarteInfo] = []

    def __init__(self, node: MoeraNode, client_scope: List[Scope] | None = None, admin_scope: List[Scope] | None = None,
                 target_node_name: str | None = None):
        """
        :param node: node to get cartes from
        :param client_scope: permissions to be granted to the cartes; if not set, all permissions of the cartes' owner
               are granted
        :param admin_scope: additional administrative permissions (of those granted to the cartes' owner by the target
               node) to be granted to the cartes
        :param target_node_name: if set, the cartes are valid for authentication on the specified node only
        """
        self._node = node
        self._client_scope = client_scope if client_scope is not None else ["all"]
        self._admin_scope = admin_scope if admin_scope is not None else []
        self._target_node_name = target_node_name

    def renew(self) -> None:
        """
        Force renewing the cached list of cartes.
        """
        attributes = CarteAttributes(client_scope=self._client_scope, admin_scope=self._admin_scope,
                                     node_name=self._target_node_name)
        self._cartes = self._node.create_cartes(attributes).cartes

    def get_carte(self) -> str:
        """
        Get a valid carte. Use one of the cached ones, if possible.

        :return: the carte
        """
        for renewed in [False, True]:
            now = int(time())
            self._cartes = [c for c in self._cartes if c.deadline > now]
            if len(self._cartes) == 0:
                if renewed:
                    raise MoeraCartesError("Could not obtain a valid carte from the node")
                self.renew()
                continue
            for c in self._cartes:
                if c.beginning <= now:
                    return c.carte
            raise MoeraCartesError("Could not obtain a carte valid for now")


def to_scope_mask(scope: List[Scope]) -> int:
    mask = 0
    for sc in scope:
        mask |= SCOPE_VALUES[sc]
    return mask


def generate_carte(owner_name: str, signing_key: ec.EllipticCurvePrivateKey, beginning: Timestamp, ttl: int = 600,
                   address: str | None = None, node_name: str | None = None,
                   client_scope: List[Scope] | int = SCOPE_VALUES["all"], admin_scope: List[Scope] | int = 0) -> str:
    """
    Generate a carte with the given parameters and sign it with the provided private signing key.

    :param owner_name: name of the node authenticating with the carte
    :param signing_key: the private signing key to sign the carte
    :param beginning: timestamp of the beginning of the carte's life
    :param ttl: length of the carte's life, in seconds
    :param address: if set, the carte is valid for authentication from the given IP address only
    :param node_name: if set, the carte is valid for authentication on the specified node only
    :param client_scope: list of permissions granted to the carte
    :param admin_scope: list of additional administrative permissions (of those granted to the carte's owner by
           the target node) granted to the carte
    :return: the carte
    """
    if isinstance(client_scope, list):
        client_scope = to_scope_mask(client_scope)
    if isinstance(admin_scope, list):
        admin_scope = to_scope_mask(admin_scope)
    fingerprint = create_carte_fingerprint2(owner_name, address, beginning, beginning + ttl, node_name, client_scope,
                                            admin_scope, os.urandom(8))
    signature = sign_fingerprint(fingerprint, signing_key)
    return urlsafe_b64encode(fingerprint + signature).decode('ascii')



from typing import List, Union, Optional, Tuple, Any, Literal, Mapping, cast

FingerprintPrimitiveType = Union[Literal['string', 'boolean', 'number', 'bytes', 'bytes[]'] | 'FingerprintSchema']
FingerprintSchema = List[Tuple[str, FingerprintPrimitiveType]]
Fingerprint = Mapping[str, Any]


class FingerprintWriter:

    def __init__(self):
        self._data = bytearray()

    def _append_null(self) -> None:
        self._data.append(0xff)

    def _append_string(self, string: Optional[str]) -> None:
        if string is None:
            self._append_null()
            return
        buf = string.encode()
        self._append_number(len(buf))
        self._data.extend(buf)

    def _append_boolean(self, b: bool) -> None:
        self._data.append(1 if b else 0)

    def _append_number(self, l: int) -> None:
        if l < 0xfc:
            len = 1
        elif l <= 0xffff:
            self._data.append(0xfc)
            len = 2
        elif l <= 0xffffffff:
            self._data.append(0xfd)
            len = 4
        else:
            self._data.append(0xfe)
            len = 8
        for _ in range(len):
            self._data.append(l & 0xff)
            l >>= 8

    def _append_bytes(self, b: bytes) -> None:
        self._append_number(len(b))
        self._data.extend(b)

    def _append_fingerprint(self, fingerprint: Fingerprint, schema: FingerprintSchema) -> None:
        for field in schema:
            self.append(fingerprint[field[0]], field[1])

    def _append_list(self, list: List[Any], type: FingerprintPrimitiveType) -> None:
        writer = FingerprintWriter()
        for value in list:
            writer.append(value, type)
        self._append_bytes(writer.to_bytes())

    def append(self, value: Any, type: FingerprintPrimitiveType) -> None:
        if value is None:
            self._append_null()
        elif isinstance(type, list):
            self._append_fingerprint(value, type)
        elif type.endswith('[]'):
            self._append_list(value, cast(FingerprintPrimitiveType, type.removesuffix('[]')))
        elif type == 'string':
            self._append_string(value)
        elif type == 'boolean':
            self._append_boolean(value)
        elif type == 'number':
            self._append_number(value)
        elif type == 'bytes':
            self._append_bytes(value)

    def to_bytes(self) -> bytes:
        return bytes(self._data)



from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from ecpy.curves import Curve
from mnemonic import Mnemonic

from .fingerprint import Fingerprint, FingerprintSchema, FingerprintWriter


def generate_key() -> ec.EllipticCurvePrivateKey:
    """
    Generate a pair of cryptographic keys.

    :return: the keys
    """
    return ec.generate_private_key(ec.SECP256K1())


def generate_mnemonic_key() -> (str, ec.EllipticCurvePrivateKey):
    """
    Generate a pair of cryptographic keys with a mnemonic for the private key.

    :return: the mnemonic and the keys
    """
    mnemonic = Mnemonic().generate(strength=256)
    private_key = mnemonic_to_private_key(mnemonic)
    return mnemonic, private_key


def mnemonic_to_private_key(mnemonic: str) -> ec.EllipticCurvePrivateKey:
    """
    Restore a private key from the given mnemonic.

    :param mnemonic: the mnemonic
    :return: the private key
    """
    seed = Mnemonic().to_seed(mnemonic)
    private_value = int.from_bytes(seed, 'big') % Curve.get_curve('secp256k1').field
    private_key = ec.derive_private_key(private_value, ec.SECP256K1())
    return private_key


def raw_public_key(public_key: ec.EllipticCurvePublicKey) -> bytes:
    """
    Convert a public key to the raw format used by the naming server.

    :param public_key: the public key
    :return: the raw public key
    """
    numbers = public_key.public_numbers()
    return numbers.x.to_bytes(32, 'big') + numbers.y.to_bytes(32, 'big')


def raw_to_public_key(raw_public_key: bytes) -> ec.EllipticCurvePublicKey:
    """
    Restore a public key from the raw format.

    :param raw_public_key: the raw public key
    :return: the public key
    """
    x = int.from_bytes(raw_public_key[:32], 'big')
    y = int.from_bytes(raw_public_key[32:], 'big')
    return ec.EllipticCurvePublicNumbers(x, y, ec.SECP256K1()).public_key()


def raw_private_key(private_key: ec.EllipticCurvePrivateKey) -> bytes:
    """
    Convert a private key to the raw format to pass to the client.

    :param private_key: the private key
    :return: the raw private key
    """
    return private_key.private_numbers().private_value.to_bytes(32, 'big')


def raw_to_private_key(raw_private_key: bytes) -> ec.EllipticCurvePrivateKey:
    """
    Restore a private key from the raw format.

    :param raw_private_key: the raw private key
    :return: the private key
    """
    return ec.derive_private_key(int.from_bytes(raw_private_key, 'big'), ec.SECP256K1())


def fingerprint_bytes(fingerprint: Fingerprint, schema: FingerprintSchema) -> bytes:
    """
    Encode a fingerprint in the binary form, using the given fingerprint data and schema.

    :param fingerprint: the fingerprint data
    :param schema: the fingerprint schema
    :return: the fingerprint in the binary form
    """
    writer = FingerprintWriter()
    writer.append(fingerprint, schema)
    return writer.to_bytes()


def digest_fingerprint(fingerprint: bytes) -> bytes:
    """
    Calculate a cryptographic digest of the fingerprint.

    :param fingerprint: the fingerprint
    :return: the digest
    """
    digest = hashes.Hash(hashes.SHA3_256())
    digest.update(fingerprint)
    return digest.finalize()


def sign_fingerprint(fingerprint: bytes, private_key: ec.EllipticCurvePrivateKey) -> bytes:
    """
    Sign a fingerprint with a private key.

    :param fingerprint: the fingerprint to be signed
    :param private_key: the private key
    :return: the signature
    """
    return private_key.sign(fingerprint, ec.ECDSA(hashes.SHA3_256()))


def verify_fingerprint_signature(fingerprint: bytes, signature: bytes, public_key: ec.EllipticCurvePublicKey) -> bool:
    """
    Verify a fingerprint signature with the given public key.

    :param fingerprint: the original fingerprint
    :param signature: the signature to be verified
    :param public_key: the public key for verification
    :return: `True`, if the signature is correct, `False` otherwise
    """
    try:
        public_key.verify(signature, fingerprint, ec.ECDSA(hashes.SHA3_256()))
        return True
    except InvalidSignature:
        return False



# This file is generated

from typing import IO, List, cast
from urllib.parse import quote_plus

from . import types, schemas
from .caller import Caller
from ..structure import comma_separated_flags, structure_list


class MoeraNode(Caller):
    """Node API interface."""

    def __init__(self, node_url: str | None = None):
        if node_url is not None:
            self.node_url(node_url)

    def search_activity_reactions(self, filter: types.ActivityReactionFilter) -> List[types.ActivityReactionInfo]:
        """
        Get the list of all reactions performed by the node, filtered by some criteria.

        :param filter:
        """
        location = "/activity/reactions/search"
        data = self.call(
            "search_activity_reactions", location, method="POST", body=filter,
            schema=schemas.ACTIVITY_REACTION_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.ActivityReactionInfo)

    def get_search_history(
        self, prefix: str | None = None, limit: int | None = None
    ) -> List[types.SearchHistoryInfo]:
        """
        Get a list of previously executed search queries, optionally filtered by the given ``prefix`` and limited by
        the given ``limit``. The node may decide to return fewer queries than the given ``limit``. The queries are
        always sorted by creation timestamp, descending.

        :param prefix: find queries with the specified prefix (case-insensitive)
        :param limit: maximum number of queries returned
        """
        location = "/activity/search".format()
        params = {"prefix": prefix, "limit": limit}
        data = self.call(
            "get_search_history", location, method="GET", params=params,
            schema=schemas.SEARCH_HISTORY_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.SearchHistoryInfo)

    def save_to_search_history(self, history_text: types.SearchHistoryText) -> types.SearchHistoryInfo:
        """
        Save a search query in the registry.

        :param history_text:
        """
        location = "/activity/search"
        data = self.call(
            "save_to_search_history", location, method="POST", body=history_text,
            schema=schemas.SEARCH_HISTORY_INFO_SCHEMA
        )
        return types.SearchHistoryInfo.from_json(data)

    def delete_from_search_history(self, query: str) -> types.Result:
        """
        Delete a search query from the registry.

        :param query: the query to be deleted
        """
        location = "/activity/search".format()
        params = {"query": query}
        data = self.call(
            "delete_from_search_history", location, method="DELETE", params=params, schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def get_remote_sheriff_orders_slice(
        self, after: int | None = None, before: int | None = None, limit: int | None = None
    ) -> types.SheriffOrdersSliceInfo:
        """
        Get a slice of the list of all orders sent by the sheriff, delimited by the ``before`` or ``after`` moment and
        the given ``limit``. If neither ``before`` nor ``after`` are provided, the latest orders are returned. The node
        may decide to return fewer orders than the given ``limit``. The orders are always sorted by moment, descending.

        :param after: filter orders posted strongly after this moment
        :param before: filter orders posted at or before this moment
        :param limit: maximum number of orders returned
        """
        location = "/activity/sheriff/orders".format()
        params = {"after": after, "before": before, "limit": limit}
        data = self.call(
            "get_remote_sheriff_orders_slice", location, method="GET", auth=False, params=params,
            schema=schemas.SHERIFF_ORDERS_SLICE_INFO_SCHEMA
        )
        return types.SheriffOrdersSliceInfo.from_json(data)

    def get_remote_posting_verification_status(self, id: str) -> types.RemotePostingVerificationInfo:
        """
        Get the status of the asynchronous operation that performs verification of a remote posting signature.

        :param id: asynchronous operation ID
        """
        location = "/async-operations/remote-posting-verification/{id}".format(id=quote_plus(id))
        data = self.call(
            "get_remote_posting_verification_status", location, method="GET",
            schema=schemas.REMOTE_POSTING_VERIFICATION_INFO_SCHEMA
        )
        return types.RemotePostingVerificationInfo.from_json(data)

    def get_remote_reaction_verification_status(self, id: str) -> types.RemoteReactionVerificationInfo:
        """
        Get the status of the asynchronous operation that performs verification of the signature of a reaction to a
        remote posting.

        :param id: asynchronous operation ID
        """
        location = "/async-operations/remote-reaction-verification/{id}".format(id=quote_plus(id))
        data = self.call(
            "get_remote_reaction_verification_status", location, method="GET",
            schema=schemas.REMOTE_REACTION_VERIFICATION_INFO_SCHEMA
        )
        return types.RemoteReactionVerificationInfo.from_json(data)

    def get_avatars(self) -> List[types.AvatarInfo]:
        """
        Get the list of avatars in the ascending order of their ordinals.
        """
        location = "/avatars"
        data = self.call("get_avatars", location, method="GET", auth=False, schema=schemas.AVATAR_INFO_ARRAY_SCHEMA)
        return structure_list(data, types.AvatarInfo)

    def create_avatar(self, avatar: types.AvatarAttributes) -> types.AvatarInfo:
        """
        Create a new avatar from a public media file that exists on the node. A new public media file is created for
        the avatar. If the avatar's ordinal is not provided in the input, the avatar is assigned an ordinal that is
        greater than ordinals of all existing avatars.

        :param avatar:
        """
        location = "/avatars"
        data = self.call("create_avatar", location, method="POST", body=avatar, schema=schemas.AVATAR_INFO_SCHEMA)
        return types.AvatarInfo.from_json(data)

    def get_avatar(self, id: str) -> types.AvatarInfo:
        """
        Get an individual avatar.

        :param id: avatar ID
        """
        location = "/avatars/{id}".format(id=quote_plus(id))
        data = self.call("get_avatar", location, method="GET", auth=False, schema=schemas.AVATAR_INFO_SCHEMA)
        return types.AvatarInfo.from_json(data)

    def delete_avatar(self, id: str) -> types.Result:
        """
        Delete an avatar.

        :param id: avatar ID
        """
        location = "/avatars/{id}".format(id=quote_plus(id))
        data = self.call("delete_avatar", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def reorder_avatars(self, order: types.AvatarsOrdered) -> List[types.AvatarOrdinal]:
        """
        Reorder avatars. Every avatar mentioned in the input is assigned an ordinal in ascending order as they appear
        in the input. Ordinals of avatars not mentioned in the input are not touched.

        :param order:
        """
        location = "/avatars/reorder"
        data = self.call(
            "reorder_avatars", location, method="POST", body=order, schema=schemas.AVATAR_ORDINAL_ARRAY_SCHEMA
        )
        return structure_list(data, types.AvatarOrdinal)

    def block_instant(self, instant: types.BlockedInstantAttributes) -> types.BlockedInstantInfo:
        """
        Blocks creation of instants of the given story type, related to the given entry, optionally unblocking at the
        given time in the future.

        :param instant:
        """
        location = "/blocked-instants"
        data = self.call(
            "block_instant", location, method="POST", body=instant, schema=schemas.BLOCKED_INSTANT_INFO_SCHEMA
        )
        return types.BlockedInstantInfo.from_json(data)

    def get_blocked_instant(self, id: str) -> types.BlockedInstantInfo:
        """
        Get details about the given blocked instant.

        :param id: ID of the blocked instant
        """
        location = "/blocked-instants/{id}".format(id=quote_plus(id))
        data = self.call("get_blocked_instant", location, method="GET", schema=schemas.BLOCKED_INSTANT_INFO_SCHEMA)
        return types.BlockedInstantInfo.from_json(data)

    def unblock_instant(self, id: str) -> types.Result:
        """
        Unblock the given instant.

        :param id: ID of the blocked instant
        """
        location = "/blocked-instants/{id}".format(id=quote_plus(id))
        data = self.call("unblock_instant", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def search_blocked_instants(self, filter: types.BlockedInstantFilter) -> List[types.BlockedInstantInfo]:
        """
        Search blocked instants by the given criteria.

        :param filter:
        """
        location = "/blocked-instants/search"
        data = self.call(
            "search_blocked_instants", location, method="POST", body=filter,
            schema=schemas.BLOCKED_INSTANT_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.BlockedInstantInfo)

    def block_user(self, user: types.BlockedUserAttributes) -> types.BlockedUserInfo:
        """
        Blocks the given node from performing the given operations, in a particular posting or globally, optionally
        unblocking at the given time in the future.

        :param user:
        """
        location = "/people/blocked-users"
        data = self.call("block_user", location, method="POST", body=user, schema=schemas.BLOCKED_USER_INFO_SCHEMA)
        return types.BlockedUserInfo.from_json(data)

    def get_blocked_user(self, id: str) -> types.BlockedUserInfo:
        """
        Get details about the given blocked user.

        :param id: ID of the blocked user
        """
        location = "/people/blocked-users/{id}".format(id=quote_plus(id))
        data = self.call("get_blocked_user", location, method="GET", schema=schemas.BLOCKED_USER_INFO_SCHEMA)
        return types.BlockedUserInfo.from_json(data)

    def unblock_user(self, id: str) -> types.Result:
        """
        Unblock the given user.

        :param id: ID of the blocked user
        """
        location = "/people/blocked-users/{id}".format(id=quote_plus(id))
        data = self.call("unblock_user", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def search_blocked_users(self, filter: types.BlockedUserFilter) -> List[types.BlockedUserInfo]:
        """
        Search blocked users by the given criteria.

        :param filter:
        """
        location = "/people/blocked-users/search"
        data = self.call(
            "search_blocked_users", location, method="POST", body=filter, schema=schemas.BLOCKED_USER_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.BlockedUserInfo)

    def get_blocked_users_checksums(self) -> types.BlockedUsersChecksums:
        """
        Get checksums of the information about the blocked users. This request may be used to quickly detect the
        changes in the blocked users list to update the cache on the client side.
        """
        location = "/people/blocked-users/checksums"
        data = self.call(
            "get_blocked_users_checksums", location, method="GET", schema=schemas.BLOCKED_USERS_CHECKSUMS_SCHEMA
        )
        return types.BlockedUsersChecksums.from_json(data)

    def get_blocked_by_user(self, id: str) -> types.BlockedByUserInfo:
        """
        Get details about the given node that blocked this node.

        :param id: ID of the blocked-by user
        """
        location = "/people/blocked-by-users/{id}".format(id=quote_plus(id))
        data = self.call("get_blocked_by_user", location, method="GET", schema=schemas.BLOCKED_BY_USER_INFO_SCHEMA)
        return types.BlockedByUserInfo.from_json(data)

    def search_blocked_by_users(self, filter: types.BlockedByUserFilter) -> List[types.BlockedByUserInfo]:
        """
        Search nodes that blocked this node, by the given criteria.

        :param filter:
        """
        location = "/people/blocked-by-users/search"
        data = self.call(
            "search_blocked_by_users", location, method="POST", body=filter,
            schema=schemas.BLOCKED_BY_USER_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.BlockedByUserInfo)

    def create_cartes(self, attributes: types.CarteAttributes) -> types.CarteSet:
        """
        Create a set of cartes with the given attributes. Cartes in the sequence correspond to successive periods of
        time.

        :param attributes:
        """
        location = "/cartes"
        data = self.call("create_cartes", location, method="POST", body=attributes, schema=schemas.CARTE_SET_SCHEMA)
        return types.CarteSet.from_json(data)

    def verify_carte(self, client_carte: types.ClientCarte) -> types.CarteVerificationInfo:
        """
        Verify if the given carte may be used for authentication on this node. Additionally, if ``clientName`` is
        provided, it is compared to the carte owner's name.

        :param client_carte:
        """
        location = "/cartes/verify"
        data = self.call(
            "verify_carte", location, method="POST", body=client_carte, schema=schemas.CARTE_VERIFICATION_INFO_SCHEMA
        )
        return types.CarteVerificationInfo.from_json(data)

    def get_comments_slice(
        self, posting_id: str, after: int | None = None, before: int | None = None, limit: int | None = None
    ) -> types.CommentsSliceInfo:
        """
        Get a slice of the list of comments, delimited by ``before`` or ``after`` moments (but not both) and the given
        ``limit``. If neither ``before`` nor ``after`` are provided, the latest comments are returned. The node may
        decide to return fewer comments than the given ``limit``. The stories are always sorted by moment, ascending.

        :param posting_id: ID of the posting
        :param after: filter comments posted strongly after this moment
        :param before: filter comments posted at or before this moment
        :param limit: maximum number of comments returned
        """
        location = "/postings/{postingId}/comments".format(postingId=quote_plus(posting_id))
        params = {"after": after, "before": before, "limit": limit}
        data = self.call(
            "get_comments_slice", location, method="GET", params=params, bodies=True,
            schema=schemas.COMMENTS_SLICE_INFO_SCHEMA
        )
        return types.CommentsSliceInfo.from_json(data)

    def create_comment(self, posting_id: str, comment: types.CommentText) -> types.CommentCreated:
        """
        Create a comment from the given text and add it to the given posting. The comment owner must authenticate in
        some way. If the comment is not signed, it will be kept for a limited period of time and then erased. If
        authenticated as admin, the node signs the comment.

        :param posting_id: ID of the posting
        :param comment:
        """
        location = "/postings/{postingId}/comments".format(postingId=quote_plus(posting_id))
        data = self.call(
            "create_comment", location, method="POST", src_bodies=True, body=comment, bodies=True,
            schema=schemas.COMMENT_CREATED_SCHEMA
        )
        return types.CommentCreated.from_json(data)

    def get_comment(self, posting_id: str, comment_id: str, with_source: bool = False) -> types.CommentInfo:
        """
        Get an individual comment.

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        :param with_source: include source text of the comment
        """
        location = "/postings/{postingId}/comments/{commentId}".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        include = comma_separated_flags({"source": with_source})
        params = {"include": include}
        data = self.call(
            "get_comment", location, method="GET", params=params, bodies=True, schema=schemas.COMMENT_INFO_SCHEMA
        )
        return types.CommentInfo.from_json(data)

    def update_all_comments(self, posting_id: str, attributes: types.CommentMassAttributes) -> types.Result:
        """
        Update operation overrides for all comments in the posting.

        :param posting_id: ID of the posting
        :param attributes:
        """
        location = "/postings/{postingId}/comments".format(postingId=quote_plus(posting_id))
        data = self.call("update_all_comments", location, method="PUT", body=attributes, schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def update_comment(self, posting_id: str, comment_id: str, comment: types.CommentText) -> types.CommentInfo:
        """
        Update the comment, creating a new revision of it. The text is processed just like in the ``POST`` request.

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        :param comment:
        """
        location = "/postings/{postingId}/comments/{commentId}".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        data = self.call(
            "update_comment", location, method="PUT", src_bodies=True, body=comment, bodies=True,
            schema=schemas.COMMENT_INFO_SCHEMA
        )
        return types.CommentInfo.from_json(data)

    def delete_comment(self, posting_id: str, comment_id: str) -> types.CommentTotalInfo:
        """
        Delete the comment. The comment may not be purged from the database immediately, but preserved for some period
        of time to give a chance to restore it.

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        """
        location = "/postings/{postingId}/comments/{commentId}".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        data = self.call("delete_comment", location, method="DELETE", schema=schemas.COMMENT_TOTAL_INFO_SCHEMA)
        return types.CommentTotalInfo.from_json(data)

    def get_postings_attached_to_comment(self, posting_id: str, comment_id: str) -> List[types.PostingInfo]:
        """
        Get all postings linked to media attached to the given comment.

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        """
        location = "/postings/{postingId}/comments/{commentId}/attached".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        data = self.call(
            "get_postings_attached_to_comment", location, method="GET", bodies=True,
            schema=schemas.POSTING_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.PostingInfo)

    def get_comment_revisions(self, posting_id: str, comment_id: str) -> List[types.CommentRevisionInfo]:
        """
        Get all revisions of the comment.

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        """
        location = "/postings/{postingId}/comments/{commentId}/revisions".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        data = self.call(
            "get_comment_revisions", location, method="GET", bodies=True,
            schema=schemas.COMMENT_REVISION_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.CommentRevisionInfo)

    def get_comment_revision(self, posting_id: str, comment_id: str, id: str) -> types.CommentRevisionInfo:
        """
        Get an individual revision of the comment.

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        :param id: ID of the revision
        """
        location = "/postings/{postingId}/comments/{commentId}/revisions/{id}".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id), id=quote_plus(id)
        )
        data = self.call(
            "get_comment_revision", location, method="GET", bodies=True, schema=schemas.COMMENT_REVISION_INFO_SCHEMA
        )
        return types.CommentRevisionInfo.from_json(data)

    def create_comment_reaction(
        self, posting_id: str, comment_id: str, reaction: types.ReactionDescription
    ) -> types.ReactionCreated:
        """
        Add a reaction to the given comment. The reaction owner must authenticate in some way. Only one reaction is
        allowed from each owner to a particular comment. If a reaction from the same owner to this comment already
        exists, it is overwritten. If the reaction is not signed, the reaction will be kept for a limited period of
        time and then erased (the previous reaction of the same owner will be restored, if any).

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        :param reaction:
        """
        location = "/postings/{postingId}/comments/{commentId}/reactions".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        data = self.call(
            "create_comment_reaction", location, method="POST", body=reaction, schema=schemas.REACTION_CREATED_SCHEMA
        )
        return types.ReactionCreated.from_json(data)

    def update_comment_reaction(
        self, posting_id: str, comment_id: str, owner_name: str, reaction: types.ReactionOverride
    ) -> types.ReactionInfo:
        """
        Update the reaction's operations or set operations' overrides.

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        :param owner_name: reaction owner node name
        :param reaction:
        """
        location = "/postings/{postingId}/comments/{commentId}/reactions/{ownerName}".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id), ownerName=quote_plus(owner_name)
        )
        data = self.call(
            "update_comment_reaction", location, method="PUT", body=reaction, schema=schemas.REACTION_INFO_SCHEMA
        )
        return types.ReactionInfo.from_json(data)

    def get_comment_reactions_slice(
        self, posting_id: str, comment_id: str, negative: bool | None = None, emoji: int | None = None,
        before: int | None = None, limit: int | None = None
    ) -> types.ReactionsSliceInfo:
        """
        Get a slice of the list of reactions to the given comment, optionally filtered by reaction type, delimited by
        ``before`` moment and the given ``limit``. If ``before`` is not provided, the latest reactions are returned.
        The node may decide to return fewer reactions than the given ``limit``. The reactions are always sorted by
        moment, descending.

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        :param negative: ``True``, to filter negative reactions, ``False``, to filter positive ones
        :param emoji: filter by reaction code, usually interpreted by clients as emoji code point
        :param before: filter reactions created at or before this moment
        :param limit: maximum number of reactions returned
        """
        location = "/postings/{postingId}/comments/{commentId}/reactions".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        params = {"negative": negative, "emoji": emoji, "before": before, "limit": limit}
        data = self.call(
            "get_comment_reactions_slice", location, method="GET", params=params,
            schema=schemas.REACTIONS_SLICE_INFO_SCHEMA
        )
        return types.ReactionsSliceInfo.from_json(data)

    def get_comment_reaction(self, posting_id: str, comment_id: str, owner_name: str) -> types.ReactionInfo:
        """
        Get the detailed information about the reaction of the given owner to the given comment. If no reaction with
        such an owner exists, an empty structure with just ``commentId`` is returned.

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        :param owner_name: reaction owner node name
        """
        location = "/postings/{postingId}/comments/{commentId}/reactions/{ownerName}".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id), ownerName=quote_plus(owner_name)
        )
        data = self.call("get_comment_reaction", location, method="GET", schema=schemas.REACTION_INFO_SCHEMA)
        return types.ReactionInfo.from_json(data)

    def delete_all_comment_reactions(self, posting_id: str, comment_id: str) -> types.Result:
        """
        Delete all reactions to the given comment.

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        """
        location = "/postings/{postingId}/comments/{commentId}/reactions".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        data = self.call("delete_all_comment_reactions", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def delete_comment_reaction(self, posting_id: str, comment_id: str, owner_name: str) -> types.ReactionTotalsInfo:
        """
        Delete the reaction of the given owner to the given comment.

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        :param owner_name: reaction owner node name
        """
        location = "/postings/{postingId}/comments/{commentId}/reactions/{ownerName}".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id), ownerName=quote_plus(owner_name)
        )
        data = self.call(
            "delete_comment_reaction", location, method="DELETE", schema=schemas.REACTION_TOTALS_INFO_SCHEMA
        )
        return types.ReactionTotalsInfo.from_json(data)

    def get_comment_reaction_totals(self, posting_id: str, comment_id: str) -> types.ReactionTotalsInfo:
        """
        Get a summary of reactions to the comment given.

        :param posting_id: ID of the posting
        :param comment_id: ID of the comment
        """
        location = "/postings/{postingId}/comments/{commentId}/reaction-totals".format(
            postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        data = self.call(
            "get_comment_reaction_totals", location, method="GET", schema=schemas.REACTION_TOTALS_INFO_SCHEMA
        )
        return types.ReactionTotalsInfo.from_json(data)

    def get_contacts(self, query: str | None = None, limit: int | None = None) -> List[types.ContactInfo]:
        """
        Search for contacts matching the search ``query``. Every space-delimited word in the query must match
        case-insensitively a beginning of the contact's node name or a beginning of any space-delimited word in the
        contact's full name. The order of words is not significant.
        
        The node may decide to return fewer contacts than the given ``limit``.
        
        The contacts are sorted by *social distance* from the node, which depends on their subscription and friendship
        status and the number of recent reactions and comments.

        :param query: the search query
        :param limit: maximum number of contacts returned
        """
        location = "/people/contacts".format()
        params = {"query": query, "limit": limit}
        data = self.call(
            "get_contacts", location, method="GET", params=params, schema=schemas.CONTACT_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.ContactInfo)

    def fetch_contacts(self, filter: types.ContactFilter) -> List[types.ContactWithRelationships]:
        """
        Fetch the detailed information, including relationships, about the contacts matching the filter.

        :param filter:
        """
        location = "/people/contacts/fetch"
        data = self.call(
            "fetch_contacts", location, method="POST", body=filter,
            schema=schemas.CONTACT_WITH_RELATIONSHIPS_ARRAY_SCHEMA
        )
        return structure_list(data, types.ContactWithRelationships)

    def check_credentials(self) -> types.CredentialsCreated:
        """
        Check whether the credentials are initialized already.
        """
        location = "/credentials"
        data = self.call(
            "check_credentials", location, method="GET", auth=False, schema=schemas.CREDENTIALS_CREATED_SCHEMA
        )
        return types.CredentialsCreated.from_json(data)

    def create_credentials(self, credentials: types.Credentials) -> types.Result:
        """
        Initialize credentials if they are not set yet. Note that this operation can be executed without
        authentication, so this should be done as soon as possible after the node installation. Sign in is not allowed
        until the credentials are set.

        :param credentials:
        """
        location = "/credentials"
        data = self.call(
            "create_credentials", location, method="POST", body=credentials, auth=False, schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def update_credentials(self, credentials: types.CredentialsChange) -> types.Result:
        """
        Update credentials. Either old password or credentials reset token should be set in the input for the operation
        to succeed. Credentials reset token is not related to the authentication token and usually is sent to the user
        by E-mail.

        :param credentials:
        """
        location = "/credentials"
        data = self.call(
            "update_credentials", location, method="PUT", body=credentials, auth=False, schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def delete_credentials(self) -> types.Result:
        """
        Delete credentials.
        """
        location = "/credentials"
        data = self.call("delete_credentials", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def reset_credentials(self) -> types.EmailHint:
        """
        The node generates credentials reset token that is sent to the node admin by E-mail or using any other way that
        is defined for recovery of credentials. This token then may be used to change the credentials without knowing
        the password.
        """
        location = "/credentials/reset"
        data = self.call("reset_credentials", location, method="POST", auth=False, schema=schemas.EMAIL_HINT_SCHEMA)
        return types.EmailHint.from_json(data)

    def verify_credentials_reset_token(self, reset_token: types.CredentialsResetToken) -> types.VerificationInfo:
        """
        Verify the credentials reset token. If the token is valid, it may be used later to change the credentials
        without knowing the password.

        :param reset_token:
        """
        location = "/credentials/reset/verify"
        data = self.call(
            "verify_credentials_reset_token", location, method="POST", body=reset_token, auth=False,
            schema=schemas.VERIFICATION_INFO_SCHEMA
        )
        return types.VerificationInfo.from_json(data)

    def get_deleted_postings(self, page: int | None = None, limit: int | None = None) -> List[types.PostingInfo]:
        """
        Get the list of deleted postings, page by page. The node may decide to use a smaller page size than the given
        ``limit``. The postings are always sorted by the deletion timestamp, descending.

        :param page: page number, 0 by default
        :param limit: page size (maximum number of postings returned), the default is defined by the node
        """
        location = "/deleted-postings".format()
        params = {"page": page, "limit": limit}
        data = self.call(
            "get_deleted_postings", location, method="GET", params=params, bodies=True,
            schema=schemas.POSTING_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.PostingInfo)

    def get_deleted_posting(self, id: str) -> types.PostingInfo:
        """
        Get an individual deleted posting.

        :param id: ID of the posting
        """
        location = "/deleted-postings/{id}".format(id=quote_plus(id))
        data = self.call(
            "get_deleted_posting", location, method="GET", bodies=True, schema=schemas.POSTING_INFO_SCHEMA
        )
        return types.PostingInfo.from_json(data)

    def restore_deleted_posting(self, id: str) -> types.PostingInfo:
        """
        Restore a posting. A new revision is created with the same content as in the latest revision.

        :param id: ID of the posting
        """
        location = "/deleted-postings/{id}/restore".format(id=quote_plus(id))
        data = self.call(
            "restore_deleted_posting", location, method="POST", bodies=True, schema=schemas.POSTING_INFO_SCHEMA
        )
        return types.PostingInfo.from_json(data)

    def get_delete_posting_revisions(
        self, posting_id: str, limit: int | None = None
    ) -> List[types.PostingRevisionInfo]:
        """
        Get all revisions of the deleted posting, but not more than ``limit``. The node may decide to return fewer
        revisions than the given ``limit``.

        :param posting_id: ID of the posting
        :param limit: maximum number of revisions returned
        """
        location = "/deleted-postings/{postingId}/revisions".format(postingId=quote_plus(posting_id))
        params = {"limit": limit}
        data = self.call(
            "get_delete_posting_revisions", location, method="GET", params=params, bodies=True,
            schema=schemas.POSTING_REVISION_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.PostingRevisionInfo)

    def get_deleted_posting_revision(self, posting_id: str, id: str) -> types.PostingRevisionInfo:
        """
        Get an individual revision of the deleted posting.

        :param posting_id: ID of the posting
        :param id: ID of the revision
        """
        location = "/deleted-postings/{postingId}/revisions/{id}".format(
            postingId=quote_plus(posting_id), id=quote_plus(id)
        )
        data = self.call(
            "get_deleted_posting_revision", location, method="GET", bodies=True,
            schema=schemas.POSTING_REVISION_INFO_SCHEMA
        )
        return types.PostingRevisionInfo.from_json(data)

    def restore_deleted_posting_revision(self, posting_id: str, id: str) -> types.PostingRevisionInfo:
        """
        Restore a posting at a particular revision. A new revision is created with the same content as in the given
        revision.

        :param posting_id: ID of the posting
        :param id: ID of the revision
        """
        location = "/postings/{postingId}/revisions/{id}/restore".format(
            postingId=quote_plus(posting_id), id=quote_plus(id)
        )
        data = self.call(
            "restore_deleted_posting_revision", location, method="POST", bodies=True,
            schema=schemas.POSTING_REVISION_INFO_SCHEMA
        )
        return types.PostingRevisionInfo.from_json(data)

    def get_domains(self) -> List[types.DomainInfo]:
        """
        Get the list of registered domains.
        """
        location = "/domains"
        data = self.call("get_domains", location, method="GET", schema=schemas.DOMAIN_INFO_ARRAY_SCHEMA)
        return structure_list(data, types.DomainInfo)

    def get_domain(self, name: str) -> types.DomainInfo:
        """
        Get information about the domain with the given hostname. If domain registration for this server is public,
        this request does not require authentication.

        :param name: domain name
        """
        location = "/domains/{name}".format(name=quote_plus(name))
        data = self.call("get_domain", location, method="GET", schema=schemas.DOMAIN_INFO_SCHEMA)
        return types.DomainInfo.from_json(data)

    def create_domain(self, domain: types.DomainAttributes) -> types.DomainInfo:
        """
        Create a new domain with the given hostname. If ``nodeId`` is not passed, it is generated automatically. If
        domain registration for this server is public, this request does not require authentication.

        :param domain:
        """
        location = "/domains"
        data = self.call("create_domain", location, method="POST", body=domain, schema=schemas.DOMAIN_INFO_SCHEMA)
        return types.DomainInfo.from_json(data)

    def update_domain(self, name: str, domain: types.DomainAttributes) -> types.DomainInfo:
        """
        Update the domain with the given hostname. If the new hostname is not passed, the old hostname is preserved.
        (Note that you cannot pass a new name for the default hostname, because it cannot be renamed and ``_default_``
        is not a valid hostname. Skip this field if you want to update the default hostname.) If ``nodeId`` is not
        passed, it is generated automatically.

        :param name: domain's hostname
        :param domain:
        """
        location = "/domains/{name}".format(name=quote_plus(name))
        data = self.call("update_domain", location, method="PUT", body=domain, schema=schemas.DOMAIN_INFO_SCHEMA)
        return types.DomainInfo.from_json(data)

    def delete_domain(self, name: str) -> types.Result:
        """
        Delete the domain with the given hostname. This operation deletes the domain record only, the user's data
        related to the domain is preserved.

        :param name: domain name
        """
        location = "/domains/{name}".format(name=quote_plus(name))
        data = self.call("delete_domain", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def is_domain_available(self, node_name: str) -> types.DomainAvailable:
        """
        Get an available domain name recommended for the given node name. The domain name is usually chosen to be close
        to the node name in English transcription. If domain registration for this server is not public, this request
        is not accessible.

        :param node_name: node name
        """
        location = "/domains/available".format()
        params = {"nodeName": node_name}
        data = self.call(
            "is_domain_available", location, method="GET", auth=False, params=params,
            schema=schemas.DOMAIN_AVAILABLE_SCHEMA
        )
        return types.DomainAvailable.from_json(data)

    def get_drafts(
        self, draft_type: types.DraftType, node_name: str, posting_id: str | None = None, comment_id: str | None = None,
        page: int | None = None, limit: int | None = None
    ) -> List[types.DraftInfo]:
        """
        Get the list of drafts, page by page, filtered by the given criteria. The node may decide to use a smaller page
        size than the given ``limit``. The drafts are always sorted by the creation timestamp, descending.

        :param draft_type: type of the drafts
        :param node_name: name of the node the drafts are related to
        :param posting_id: ID of the posting, mandatory for all types, except ``new-posting``
        :param comment_id: ID of the comment, mandatory for ``comment-update`` type
        :param page: page number, 0 by default
        :param limit: page size (maximum number of postings returned), the default is defined by the node
        """
        location = "/drafts".format()
        params = {
            "draftType": draft_type, "nodeName": node_name, "postingId": posting_id, "commentId": comment_id,
            "page": page, "limit": limit
        }
        data = self.call(
            "get_drafts", location, method="GET", params=params, bodies=True, schema=schemas.DRAFT_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.DraftInfo)

    def create_draft(self, draft: types.DraftText) -> types.DraftInfo:
        """
        Create a new draft from the text given.

        :param draft:
        """
        location = "/drafts"
        data = self.call(
            "create_draft", location, method="POST", src_bodies=True, body=draft, bodies=True,
            schema=schemas.DRAFT_INFO_SCHEMA
        )
        return types.DraftInfo.from_json(data)

    def get_draft(self, id: str) -> types.DraftInfo:
        """
        Get an individual draft.

        :param id: ID of the draft
        """
        location = "/drafts/{id}".format(id=quote_plus(id))
        data = self.call("get_draft", location, method="GET", bodies=True, schema=schemas.DRAFT_INFO_SCHEMA)
        return types.DraftInfo.from_json(data)

    def update_draft(self, id: str, draft: types.DraftText) -> types.DraftInfo:
        """
        Update the draft.

        :param id: ID of the draft
        :param draft:
        """
        location = "/drafts/{id}".format(id=quote_plus(id))
        data = self.call(
            "update_draft", location, method="PUT", src_bodies=True, body=draft, bodies=True,
            schema=schemas.DRAFT_INFO_SCHEMA
        )
        return types.DraftInfo.from_json(data)

    def delete_draft(self, id: str) -> types.Result:
        """
        Delete the draft.

        :param id: ID of the draft
        """
        location = "/drafts/{id}".format(id=quote_plus(id))
        data = self.call("delete_draft", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def get_features(self) -> types.Features:
        """
        Get information about features supported by the node.
        """
        location = "/features"
        data = self.call("get_features", location, method="GET", schema=schemas.FEATURES_SCHEMA)
        return types.Features.from_json(data)

    def get_feeds(self) -> List[types.FeedInfo]:
        """
        Get general information about all feeds accessible by client.
        """
        location = "/feeds"
        data = self.call("get_feeds", location, method="GET", schema=schemas.FEED_INFO_ARRAY_SCHEMA)
        return structure_list(data, types.FeedInfo)

    def get_feed_general(self, feed_name: str) -> types.FeedInfo:
        """
        Get general information about the feed.

        :param feed_name: name of the feed
        """
        location = "/feeds/{feedName}".format(feedName=quote_plus(feed_name))
        data = self.call("get_feed_general", location, method="GET", schema=schemas.FEED_INFO_SCHEMA)
        return types.FeedInfo.from_json(data)

    def get_feed_status(self, feed_name: str) -> types.FeedStatus:
        """
        Get information about the total number and number of non-read and non-viewed stories in the feed.

        :param feed_name: name of the feed
        """
        location = "/feeds/{feedName}/status".format(feedName=quote_plus(feed_name))
        data = self.call("get_feed_status", location, method="GET", schema=schemas.FEED_STATUS_SCHEMA)
        return types.FeedStatus.from_json(data)

    def update_feed_status(self, feed_name: str, change: types.FeedStatusChange) -> types.FeedStatus:
        """
        Update information about non-read and non-viewed stories in the feed.

        :param feed_name: name of the feed
        :param change:
        """
        location = "/feeds/{feedName}/status".format(feedName=quote_plus(feed_name))
        data = self.call("update_feed_status", location, method="PUT", body=change, schema=schemas.FEED_STATUS_SCHEMA)
        return types.FeedStatus.from_json(data)

    def get_feed_slice(
        self, feed_name: str, after: int | None = None, before: int | None = None, limit: int | None = None
    ) -> types.FeedSliceInfo:
        """
        Get a slice of the feed, delimited by ``before`` or ``after`` moments (but not both) and the given ``limit``.
        If neither ``before`` nor ``after`` are provided, the latest stories are returned. The node may decide to
        return fewer stories than the given ``limit``. The stories are always sorted by moment, descending.

        :param feed_name: name of the feed
        :param after: filter stories posted strongly after this moment
        :param before: filter stories posted at or before this moment
        :param limit: maximum number of stories returned
        """
        location = "/feeds/{feedName}/stories".format(feedName=quote_plus(feed_name))
        params = {"after": after, "before": before, "limit": limit}
        data = self.call(
            "get_feed_slice", location, method="GET", params=params, bodies=True, schema=schemas.FEED_SLICE_INFO_SCHEMA
        )
        return types.FeedSliceInfo.from_json(data)

    def delete_feed_stories(
        self, feed_name: str, type: types.StoryType | None = None, receiver: str | None = None,
        recommended: bool | None = None
    ) -> types.Result:
        """
        Delete all stories from the feed with optional filtering.

        :param feed_name: name of the feed
        :param type: delete only the stories of the given type
        :param receiver: delete only the stories about postings located at the given node
        :param recommended: delete only the stories about recommended postings
        """
        location = "/feeds/{feedName}/stories".format(feedName=quote_plus(feed_name))
        params = {"type": type, "receiver": receiver, "recommended": recommended}
        data = self.call("delete_feed_stories", location, method="DELETE", params=params, schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def get_friend_groups(self) -> List[types.FriendGroupInfo]:
        """
        Get the list of all groups of friends that exist on the node.
        """
        location = "/people/friends/groups"
        data = self.call("get_friend_groups", location, method="GET", schema=schemas.FRIEND_GROUP_INFO_ARRAY_SCHEMA)
        return structure_list(data, types.FriendGroupInfo)

    def get_friend_group(self, id: str) -> types.FriendGroupInfo:
        """
        Get the information about the group of friends.

        :param id: ID of the group of friends
        """
        location = "/people/friends/groups/{id}".format(id=quote_plus(id))
        data = self.call("get_friend_group", location, method="GET", schema=schemas.FRIEND_GROUP_INFO_SCHEMA)
        return types.FriendGroupInfo.from_json(data)

    def create_friend_group(self, friend_group: types.FriendGroupDescription) -> types.FriendGroupInfo:
        """
        Create a group of friends.

        :param friend_group:
        """
        location = "/people/friends/groups"
        data = self.call(
            "create_friend_group", location, method="POST", body=friend_group, schema=schemas.FRIEND_GROUP_INFO_SCHEMA
        )
        return types.FriendGroupInfo.from_json(data)

    def update_friend_group(self, id: str, friend_group: types.FriendGroupDescription) -> types.FriendGroupInfo:
        """
        Update the details of the group of friends.

        :param id: ID of the group of friends
        :param friend_group:
        """
        location = "/people/friends/groups/{id}".format(id=quote_plus(id))
        data = self.call(
            "update_friend_group", location, method="PUT", body=friend_group, schema=schemas.FRIEND_GROUP_INFO_SCHEMA
        )
        return types.FriendGroupInfo.from_json(data)

    def delete_friend_group(self, id: str) -> types.Result:
        """
        Delete the group of friends.

        :param id: ID of the group of friends
        """
        location = "/people/friends/groups/{id}".format(id=quote_plus(id))
        data = self.call("delete_friend_group", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def get_friends(self, group_id: str | None = None) -> List[types.FriendInfo]:
        """
        Get the list of all friends of the node or friends belonging to a particular group.

        :param group_id: ID of a group of friends
        """
        location = "/people/friends".format()
        params = {"groupId": group_id}
        data = self.call("get_friends", location, method="GET", params=params, schema=schemas.FRIEND_INFO_ARRAY_SCHEMA)
        return structure_list(data, types.FriendInfo)

    def get_friend(self, name: str) -> types.FriendInfo:
        """
        Get the friendship information for the node given.

        :param name: name of the node
        """
        location = "/people/friends/{name}".format(name=quote_plus(name))
        data = self.call("get_friend", location, method="GET", schema=schemas.FRIEND_INFO_SCHEMA)
        return types.FriendInfo.from_json(data)

    def update_friends(self, friends: List[types.FriendDescription]) -> List[types.FriendInfo]:
        """
        Update the friendship status of the nodes passed in the input. If some node passed in the input is not a member
        of some of the groups of friends listed for it, the node is added to them. If it is a member of some groups of
        friends that are not listed for it, the node is removed from them.

        :param friends:
        """
        location = "/people/friends"
        data = self.call(
            "update_friends", location, method="PUT", body=friends, schema=schemas.FRIEND_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.FriendInfo)

    def get_friend_ofs(self) -> List[types.FriendOfInfo]:
        """
        Get the list of all nodes that added this node to their friends.
        """
        location = "/people/friend-ofs"
        data = self.call("get_friend_ofs", location, method="GET", schema=schemas.FRIEND_OF_INFO_ARRAY_SCHEMA)
        return structure_list(data, types.FriendOfInfo)

    def get_friend_of(self, name: str) -> types.FriendOfInfo:
        """
        Get the information for the node given, whether it has added this node to its friends.

        :param name: name of the node
        """
        location = "/people/friend-ofs/{name}".format(name=quote_plus(name))
        data = self.call("get_friend_of", location, method="GET", schema=schemas.FRIEND_OF_INFO_SCHEMA)
        return types.FriendOfInfo.from_json(data)

    def get_all_grants(self) -> List[types.GrantInfo]:
        """
        Get the list of all nodes having administrative permissions on this node.
        """
        location = "/grants"
        data = self.call("get_all_grants", location, method="GET", schema=schemas.GRANT_INFO_ARRAY_SCHEMA)
        return structure_list(data, types.GrantInfo)

    def get_grant(self, node_name: str) -> types.GrantInfo:
        """
        Get information about the administrative permissions granted to the node.

        :param node_name: name of the node
        """
        location = "/grants/{nodeName}".format(nodeName=quote_plus(node_name))
        data = self.call("get_grant", location, method="GET", schema=schemas.GRANT_INFO_SCHEMA)
        return types.GrantInfo.from_json(data)

    def grant_or_revoke(self, node_name: str, change: types.GrantChange) -> types.GrantInfo:
        """
        Grant a set of administrative permissions to the node or revoke them.

        :param node_name: name of the node
        :param change:
        """
        location = "/grants/{nodeName}".format(nodeName=quote_plus(node_name))
        data = self.call("grant_or_revoke", location, method="PUT", body=change, schema=schemas.GRANT_INFO_SCHEMA)
        return types.GrantInfo.from_json(data)

    def revoke_all(self, node_name: str) -> types.Result:
        """
        Revoke all administrative permissions granted to the node.

        :param node_name: name of the node
        """
        location = "/grants/{nodeName}".format(nodeName=quote_plus(node_name))
        data = self.call("revoke_all", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def upload_admin_media(self, file: IO, file_type: str) -> types.PrivateMediaFileInfo:
        """
        Upload a new media file owned by the node admin. The content of the file is passed in the request body.

        :param file:
        :param file_type: content-type of ``file``
        """
        location = "/media/private"
        data = self.call(
            "upload_admin_media", location, method="POST", body_file=file, body_file_type=file_type,
            schema=schemas.PRIVATE_MEDIA_FILE_INFO_SCHEMA
        )
        return types.PrivateMediaFileInfo.from_json(data)

    def upload_private_media(self, client_name: str, file: IO, file_type: str) -> types.PrivateMediaFileInfo:
        """
        Upload a new media file owned by the given node. The content of the file is passed in the request body.

        :param client_name: name of the node owning the media file
        :param file:
        :param file_type: content-type of ``file``
        """
        location = "/media/private/{clientName}".format(clientName=quote_plus(client_name))
        data = self.call(
            "upload_private_media", location, method="POST", body_file=file, body_file_type=file_type,
            schema=schemas.PRIVATE_MEDIA_FILE_INFO_SCHEMA
        )
        return types.PrivateMediaFileInfo.from_json(data)

    def get_private_media(self, id: str, width: int | None = None, download: bool | None = None) -> IO:
        """
        Get media file content (returned in the response body).

        :param id: media file ID
        :param width: preferred width of the media in pixels; if present, the node will try to return the smallest
            in size, but the best in quality variant of the media, according to the width provided
        :param download: if ``True``, the node will add ``Content-Disposition: attachment`` header to the output
        """
        location = "/media/private/{id}/data".format(id=quote_plus(id))
        params = {"width": width, "download": download}
        data = self.call("get_private_media", location, method="GET", params=params, schema="blob")
        return cast(IO, data)

    def get_private_media_info(self, id: str) -> types.PrivateMediaFileInfo:
        """
        Get media file details.

        :param id: media file ID
        """
        location = "/media/private/{id}/info".format(id=quote_plus(id))
        data = self.call(
            "get_private_media_info", location, method="GET", schema=schemas.PRIVATE_MEDIA_FILE_INFO_SCHEMA
        )
        return types.PrivateMediaFileInfo.from_json(data)

    def get_private_media_parent_entry(self, id: str) -> List[types.EntryInfo]:
        """
        Get the list of all postings and comments the media file is attached to.

        :param id: media file ID
        """
        location = "/media/private/{id}/parent".format(id=quote_plus(id))
        data = self.call(
            "get_private_media_parent_entry", location, method="GET", bodies=True,
            schema=schemas.ENTRY_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.EntryInfo)

    def upload_public_media(self, file: IO, file_type: str) -> types.PublicMediaFileInfo:
        """
        Upload a new media file. The content of the file is passed in the request body

        :param file:
        :param file_type: content-type of ``file``
        """
        location = "/media/public"
        data = self.call(
            "upload_public_media", location, method="POST", body_file=file, body_file_type=file_type,
            schema=schemas.PUBLIC_MEDIA_FILE_INFO_SCHEMA
        )
        return types.PublicMediaFileInfo.from_json(data)

    def get_public_media(self, id: str, width: int | None = None, download: bool | None = None) -> IO:
        """
        Get media file content (returned in the response body).

        :param id: media file ID
        :param width: preferred width of the media in pixels; if present, the node will try to return the smallest
            in size, but the best in quality variant of the media, according to the width provided
        :param download: if ``True``, the node will add ``Content-Disposition: attachment`` header to the output
        """
        location = "/media/public/{id}/data".format(id=quote_plus(id))
        params = {"width": width, "download": download}
        data = self.call("get_public_media", location, method="GET", auth=False, params=params, schema="blob")
        return cast(IO, data)

    def get_public_media_info(self, id: str) -> types.PublicMediaFileInfo:
        """
        Get media file details.

        :param id: media file ID
        """
        location = "/media/public/{id}/info".format(id=quote_plus(id))
        data = self.call(
            "get_public_media_info", location, method="GET", auth=False, schema=schemas.PUBLIC_MEDIA_FILE_INFO_SCHEMA
        )
        return types.PublicMediaFileInfo.from_json(data)

    def get_node_name(self) -> types.NodeNameInfo:
        """
        Get the name of the node. Admin user receives the current status of the latest operation with the node name.
        """
        location = "/node-name"
        data = self.call("get_node_name", location, method="GET", schema=schemas.NODE_NAME_INFO_SCHEMA)
        return types.NodeNameInfo.from_json(data)

    def create_node_name(self, name_to_register: types.NameToRegister) -> types.RegisteredNameSecret:
        """
        Register a new name for the node. The corresponding signing key is generated automatically and stored at the
        node. The updating key is generated and returned in the encoded form and in the form of mnemonic (a sequence of
        English words). The words need to be written down and stored securely to be able to perform further operations
        with the name.

        :param name_to_register:
        """
        location = "/node-name"
        data = self.call(
            "create_node_name", location, method="POST", body=name_to_register,
            schema=schemas.REGISTERED_NAME_SECRET_SCHEMA
        )
        return types.RegisteredNameSecret.from_json(data)

    def update_node_name(self, secret: types.RegisteredNameSecret) -> types.Result:
        """
        Update the name of the node. May be used to assign an already-registered name to the node (the corresponding
        signing key is generated automatically and stored at the node), or to prolong the name. The secret or mnemonic
        of the updating key must be provided for this operation.

        :param secret:
        """
        location = "/node-name"
        data = self.call("update_node_name", location, method="PUT", body=secret, schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def delete_node_name(self) -> types.Result:
        """
        Delete all the information related to the node name (including the signing key) from the node. The name record
        on the naming server is not touched.
        """
        location = "/node-name"
        data = self.call("delete_node_name", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def get_stored_mnemonic(self) -> types.KeyMnemonic:
        """
        Get the updating key mnemonic stored on the node.
        """
        location = "/node-name/mnemonic"
        data = self.call("get_stored_mnemonic", location, method="GET", schema=schemas.KEY_MNEMONIC_SCHEMA)
        return types.KeyMnemonic.from_json(data)

    def store_mnemonic(self, mnemonic: types.KeyMnemonic) -> types.Result:
        """
        Store the updating key mnemonic on the node.

        :param mnemonic:
        """
        location = "/node-name/mnemonic"
        data = self.call("store_mnemonic", location, method="POST", body=mnemonic, schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def delete_stored_mnemonic(self) -> types.Result:
        """
        Delete the updating key mnemonic stored on the node.
        """
        location = "/node-name/mnemonic"
        data = self.call("delete_stored_mnemonic", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def send_notification(self, packet: types.NotificationPacket) -> types.Result:
        """
        Accept a notification packet from another node. Notification packets older than 10 minutes are ignored. The
        sending node should update the packet timestamp and the signature and send the packet again. This mechanism
        prevents attackers from recording and resending old signed packets.

        :param packet:
        """
        location = "/notifications"
        data = self.call(
            "send_notification", location, method="POST", body=packet, auth=False, schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def get_people_general(self) -> types.PeopleGeneralInfo:
        """
        Get general information about other nodes.
        """
        location = "/people"
        data = self.call("get_people_general", location, method="GET", schema=schemas.PEOPLE_GENERAL_INFO_SCHEMA)
        return types.PeopleGeneralInfo.from_json(data)

    def register_plugin(self, plugin: types.PluginDescription) -> types.PluginInfo:
        """
        Register the plugin. If the plugin authenticates as root admin, the plugin is registered at the server level.
        If the plugin authenticates as node admin, the plugin is registered at the node level.

        :param plugin:
        """
        location = "/plugins"
        data = self.call("register_plugin", location, method="POST", body=plugin, schema=schemas.PLUGIN_INFO_SCHEMA)
        return types.PluginInfo.from_json(data)

    def get_plugins(self) -> List[types.PluginInfo]:
        """
        Get information about all plugins registered for the node and server.
        """
        location = "/plugins"
        data = self.call("get_plugins", location, method="GET", schema=schemas.PLUGIN_INFO_ARRAY_SCHEMA)
        return structure_list(data, types.PluginInfo)

    def get_plugin(self, plugin_name: str) -> types.PluginInfo:
        """
        Get information about the plugin.

        :param plugin_name: name of the plugin
        """
        location = "/plugins/{pluginName}".format(pluginName=quote_plus(plugin_name))
        data = self.call("get_plugin", location, method="GET", schema=schemas.PLUGIN_INFO_SCHEMA)
        return types.PluginInfo.from_json(data)

    def unregister_plugin(self, plugin_name: str) -> types.Result:
        """
        Unregister the plugin.

        :param plugin_name: name of the plugin
        """
        location = "/plugins/{pluginName}".format(pluginName=quote_plus(plugin_name))
        data = self.call("unregister_plugin", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def create_posting(self, posting: types.PostingText) -> types.PostingInfo:
        """
        Create a new posting from the text given and publish it in the given feeds (if any). The heading and the
        preview of the posting are created automatically, if needed. The posting owner must authenticate in some way.
        If the posting is not signed, it will be kept for a limited period of time and then erased. If authenticated as
        admin, the node signs the posting.

        :param posting:
        """
        location = "/postings"
        data = self.call(
            "create_posting", location, method="POST", src_bodies=True, body=posting, bodies=True,
            schema=schemas.POSTING_INFO_SCHEMA
        )
        return types.PostingInfo.from_json(data)

    def update_posting(self, id: str, posting: types.PostingText) -> types.PostingInfo:
        """
        Update the posting, creating a new revision of it. The text is processed just like in the ``POST`` request.

        :param id: ID of the posting
        :param posting:
        """
        location = "/postings/{id}".format(id=quote_plus(id))
        data = self.call(
            "update_posting", location, method="PUT", src_bodies=True, body=posting, bodies=True,
            schema=schemas.POSTING_INFO_SCHEMA
        )
        return types.PostingInfo.from_json(data)

    def get_posting(self, id: str, with_source: bool = False) -> types.PostingInfo:
        """
        Get an individual posting.

        :param id: ID of the posting
        :param with_source: include source text of the posting
        """
        location = "/postings/{id}".format(id=quote_plus(id))
        include = comma_separated_flags({"source": with_source})
        params = {"include": include}
        data = self.call(
            "get_posting", location, method="GET", params=params, bodies=True, schema=schemas.POSTING_INFO_SCHEMA
        )
        return types.PostingInfo.from_json(data)

    def delete_posting(self, id: str) -> types.Result:
        """
        Delete the posting. The posting may not be purged from the database immediately, but preserved for some period
        of time to give a chance to restore it.

        :param id: ID of the posting
        """
        location = "/postings/{id}".format(id=quote_plus(id))
        data = self.call("delete_posting", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def get_postings_attached_to_posting(self, id: str) -> List[types.PostingInfo]:
        """
        Get all postings linked to media attached to the given posting.

        :param id: ID of the posting
        """
        location = "/postings/{id}/attached".format(id=quote_plus(id))
        data = self.call(
            "get_postings_attached_to_posting", location, method="GET", bodies=True,
            schema=schemas.POSTING_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.PostingInfo)

    def get_posting_revisions(self, posting_id: str, limit: int | None = None) -> List[types.PostingRevisionInfo]:
        """
        Get all revisions of the posting, but not more than ``limit``. The node may decide to return fewer revisions
        than the given ``limit``.

        :param posting_id: ID of the posting
        :param limit: maximum number of revisions returned
        """
        location = "/postings/{postingId}/revisions".format(postingId=quote_plus(posting_id))
        params = {"limit": limit}
        data = self.call(
            "get_posting_revisions", location, method="GET", params=params, bodies=True,
            schema=schemas.POSTING_REVISION_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.PostingRevisionInfo)

    def get_posting_revision(self, posting_id: str, id: str) -> types.PostingRevisionInfo:
        """
        Get an individual revision of the posting.

        :param posting_id: ID of the posting
        :param id: ID of the revision
        """
        location = "/postings/{postingId}/revisions/{id}".format(postingId=quote_plus(posting_id), id=quote_plus(id))
        data = self.call(
            "get_posting_revision", location, method="GET", bodies=True, schema=schemas.POSTING_REVISION_INFO_SCHEMA
        )
        return types.PostingRevisionInfo.from_json(data)

    def restore_posting_revision(self, posting_id: str, id: str) -> types.PostingRevisionInfo:
        """
        Restore a revision of the posting. A new revision is created with the same content as in the given revision.

        :param posting_id: ID of the posting
        :param id: ID of the revision
        """
        location = "/postings/{postingId}/revisions/{id}/restore".format(
            postingId=quote_plus(posting_id), id=quote_plus(id)
        )
        data = self.call(
            "restore_posting_revision", location, method="POST", bodies=True,
            schema=schemas.POSTING_REVISION_INFO_SCHEMA
        )
        return types.PostingRevisionInfo.from_json(data)

    def create_posting_reaction(self, posting_id: str, reaction: types.ReactionDescription) -> types.ReactionCreated:
        """
        Add a reaction to the given posting. The reaction owner must authenticate in some way. Only one reaction is
        allowed from each owner to a particular posting. If a reaction from the same owner to this posting already
        exists, it is overwritten. If the reaction is not signed, the reaction will be kept for a limited period of
        time and then erased (the previous reaction of the same owner will be restored, if any).

        :param posting_id: ID of the posting
        :param reaction:
        """
        location = "/postings/{postingId}/reactions".format(postingId=quote_plus(posting_id))
        data = self.call(
            "create_posting_reaction", location, method="POST", body=reaction, schema=schemas.REACTION_CREATED_SCHEMA
        )
        return types.ReactionCreated.from_json(data)

    def get_posting_reactions_slice(
        self, posting_id: str, negative: bool | None = None, emoji: int | None = None, before: int | None = None,
        limit: int | None = None
    ) -> types.ReactionsSliceInfo:
        """
        Get a slice of the list of reactions to the given posting, optionally filtered by reaction type, delimited by
        ``before`` moment and the given ``limit``. If ``before`` is not provided, the latest reactions are returned.
        The node may decide to return fewer reactions than the given ``limit``. The reactions are always sorted by
        moment, descending.

        :param posting_id: ID of the posting
        :param negative: ``True``, to filter negative reactions, ``False``, to filter positive ones
        :param emoji: filter by reaction code, usually interpreted by clients as emoji code point
        :param before: filter reactions created at or before this moment
        :param limit: maximum number of reactions returned
        """
        location = "/postings/{postingId}/reactions".format(postingId=quote_plus(posting_id))
        params = {"negative": negative, "emoji": emoji, "before": before, "limit": limit}
        data = self.call(
            "get_posting_reactions_slice", location, method="GET", params=params,
            schema=schemas.REACTIONS_SLICE_INFO_SCHEMA
        )
        return types.ReactionsSliceInfo.from_json(data)

    def update_posting_reaction(
        self, posting_id: str, owner_name: str, reaction: types.ReactionOverride
    ) -> types.ReactionInfo:
        """
        Update the reaction's operations or set operations' overrides.

        :param posting_id: ID of the posting
        :param owner_name: reaction owner node name
        :param reaction:
        """
        location = "/postings/{postingId}/reactions/{ownerName}".format(
            postingId=quote_plus(posting_id), ownerName=quote_plus(owner_name)
        )
        data = self.call(
            "update_posting_reaction", location, method="PUT", body=reaction, schema=schemas.REACTION_INFO_SCHEMA
        )
        return types.ReactionInfo.from_json(data)

    def get_posting_reaction(self, posting_id: str, owner_name: str) -> types.ReactionInfo:
        """
        Get the detailed information about the reaction of the given owner to the given posting. If no reaction with
        such an owner exists, an empty structure with just ``postingId`` is returned.

        :param posting_id: ID of the posting
        :param owner_name: reaction owner node name
        """
        location = "/postings/{postingId}/reactions/{ownerName}".format(
            postingId=quote_plus(posting_id), ownerName=quote_plus(owner_name)
        )
        data = self.call("get_posting_reaction", location, method="GET", schema=schemas.REACTION_INFO_SCHEMA)
        return types.ReactionInfo.from_json(data)

    def delete_all_posting_reactions(self, posting_id: str) -> types.Result:
        """
        Delete all reactions to the given posting.

        :param posting_id: ID of the posting
        """
        location = "/postings/{postingId}/reactions".format(postingId=quote_plus(posting_id))
        data = self.call("delete_all_posting_reactions", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def delete_posting_reaction(self, posting_id: str, owner_name: str) -> types.ReactionTotalsInfo:
        """
        Delete the reaction of the given owner to the given posting.

        :param posting_id: ID of the posting
        :param owner_name: reaction owner node name
        """
        location = "/postings/{postingId}/reactions/{ownerName}".format(
            postingId=quote_plus(posting_id), ownerName=quote_plus(owner_name)
        )
        data = self.call(
            "delete_posting_reaction", location, method="DELETE", schema=schemas.REACTION_TOTALS_INFO_SCHEMA
        )
        return types.ReactionTotalsInfo.from_json(data)

    def search_posting_reactions(self, filter: types.ReactionsFilter) -> List[types.ReactionInfo]:
        """
        Search reactions by criteria provided. Both reaction owner and at least one posting ID should be provided to
        search, otherwise an empty list is returned.

        :param filter:
        """
        location = "/postings/reactions/search"
        data = self.call(
            "search_posting_reactions", location, method="POST", body=filter, schema=schemas.REACTION_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.ReactionInfo)

    def get_posting_reaction_totals(self, posting_id: str) -> types.ReactionTotalsInfo:
        """
        Get a summary of reactions to the posting given.

        :param posting_id: ID of the posting
        """
        location = "/postings/{postingId}/reaction-totals".format(postingId=quote_plus(posting_id))
        data = self.call(
            "get_posting_reaction_totals", location, method="GET", schema=schemas.REACTION_TOTALS_INFO_SCHEMA
        )
        return types.ReactionTotalsInfo.from_json(data)

    def search_posting_reaction_totals(self, filter: types.ReactionTotalsFilter) -> List[types.ReactionTotalsInfo]:
        """
        Search summaries of reactions by criteria provided. At least one posting ID should be provided to search,
        otherwise an empty list is returned.

        :param filter:
        """
        location = "/postings/reaction-totals/search"
        data = self.call(
            "search_posting_reaction_totals", location, method="POST", body=filter,
            schema=schemas.REACTION_TOTALS_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.ReactionTotalsInfo)

    def get_profile(self, with_source: bool = False) -> types.ProfileInfo:
        """
        Get the profile.

        :param with_source: include source text of the bio
        """
        location = "/profile".format()
        include = comma_separated_flags({"source": with_source})
        params = {"include": include}
        data = self.call("get_profile", location, method="GET", params=params, schema=schemas.PROFILE_INFO_SCHEMA)
        return types.ProfileInfo.from_json(data)

    def update_profile(self, profile: types.ProfileAttributes) -> types.ProfileInfo:
        """
        Update the profile. Fields that are not set in the request body are left intact. Fields that are set to an
        empty value are reset to their defaults.

        :param profile:
        """
        location = "/profile"
        data = self.call("update_profile", location, method="PUT", body=profile, schema=schemas.PROFILE_INFO_SCHEMA)
        return types.ProfileInfo.from_json(data)

    def verify_email(self) -> types.Result:
        """
        Repeat verification process (resend the confirmation mail) for the current e-mail address set in the profile.
        """
        location = "/profile/email/verify"
        data = self.call("verify_email", location, method="POST", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def get_delete_node_request_status(self) -> types.DeleteNodeStatus:
        """
        Get the current status of the request to delete the node.
        """
        location = "/provider/delete-node"
        data = self.call(
            "get_delete_node_request_status", location, method="GET", schema=schemas.DELETE_NODE_STATUS_SCHEMA
        )
        return types.DeleteNodeStatus.from_json(data)

    def send_delete_node_request(self, delete_node_text: types.DeleteNodeText) -> types.DeleteNodeStatus:
        """
        Send a request to the provider to delete the node.

        :param delete_node_text:
        """
        location = "/provider/delete-node"
        data = self.call(
            "send_delete_node_request", location, method="POST", body=delete_node_text,
            schema=schemas.DELETE_NODE_STATUS_SCHEMA
        )
        return types.DeleteNodeStatus.from_json(data)

    def cancel_delete_node_request(self) -> types.DeleteNodeStatus:
        """
        Cancel the request to delete the node.
        """
        location = "/provider/delete-node"
        data = self.call(
            "cancel_delete_node_request", location, method="DELETE", schema=schemas.DELETE_NODE_STATUS_SCHEMA
        )
        return types.DeleteNodeStatus.from_json(data)

    def proxy_media(self, url: str) -> IO:
        """
        Open the URL passed in the parameters and pass to the client the media file returned by the server.

        :param url:
        """
        location = "/proxy/media".format()
        params = {"url": url}
        data = self.call("proxy_media", location, method="GET", params=params, schema="blob")
        return cast(IO, data)

    def proxy_link_preview(self, url: str) -> types.LinkPreviewInfo:
        """
        Parse the page located at the URL and return the title, the description and the picture that may be used to
        build a preview of the page.

        :param url:
        """
        location = "/proxy/link-preview".format()
        params = {"url": url}
        data = self.call(
            "proxy_link_preview", location, method="GET", params=params, schema=schemas.LINK_PREVIEW_INFO_SCHEMA
        )
        return types.LinkPreviewInfo.from_json(data)

    def register_at_push_relay(self, attributes: types.PushRelayClientAttributes) -> types.Result:
        """
        Register a client at the push relay server to receive messages from this node. The operation is synchronous.

        :param attributes:
        """
        location = "/push-relay"
        data = self.call(
            "register_at_push_relay", location, method="POST", body=attributes, schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def get_recommended_postings(
        self, feed: str | None = None, sheriff: str | None = None, limit: int | None = None
    ) -> List[types.RecommendedPostingInfo]:
        """
        Find postings known to the recommendation service and may be of interest to the client. If the client is
        authenticated, the service may tune the recommendations for them.
        
        The service may decide to return fewer recommendations than the given ``limit``.

        :param feed: name of the feed to get recommendations for ("news" by default); recommendations for every
            feed are tracked separately
        :param sheriff: filter out entries prohibited by the given sheriff
        :param limit: maximum number of recommendations returned
        """
        location = "/recommendations/postings".format()
        params = {"feed": feed, "sheriff": sheriff, "limit": limit}
        data = self.call(
            "get_recommended_postings", location, method="GET", params=params,
            schema=schemas.RECOMMENDED_POSTING_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.RecommendedPostingInfo)

    def get_recommended_postings_for_reading(
        self, sheriff: str | None = None, limit: int | None = None
    ) -> List[types.RecommendedPostingInfo]:
        """
        Find postings known to the recommendation service and may be of interest to the client to read them. If the
        client is authenticated, the service may tune the recommendations for them.
        
        The service may decide to return fewer recommendations than the given ``limit``.

        :param sheriff: filter out entries prohibited by the given sheriff
        :param limit: maximum number of recommendations returned
        """
        location = "/recommendations/postings/reading".format()
        params = {"sheriff": sheriff, "limit": limit}
        data = self.call(
            "get_recommended_postings_for_reading", location, method="GET", params=params,
            schema=schemas.RECOMMENDED_POSTING_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.RecommendedPostingInfo)

    def get_recommended_postings_for_commenting(
        self, sheriff: str | None = None, limit: int | None = None
    ) -> List[types.RecommendedPostingInfo]:
        """
        Find postings known to the recommendation service and may be of interest to the client to take part in the
        discussion. If the client is authenticated, the service may tune the recommendations for them.
        
        The service may decide to return fewer recommendations than the given ``limit``.

        :param sheriff: filter out entries prohibited by the given sheriff
        :param limit: maximum number of recommendations returned
        """
        location = "/recommendations/postings/commenting".format()
        params = {"sheriff": sheriff, "limit": limit}
        data = self.call(
            "get_recommended_postings_for_commenting", location, method="GET", params=params,
            schema=schemas.RECOMMENDED_POSTING_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.RecommendedPostingInfo)

    def accept_recommended_posting(self, node_name: str, posting_id: str, feed: str | None = None) -> types.Result:
        """
        Inform the recommendation service that the recommended posting was accepted by the client.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        :param feed: name of the feed the recommendation is accepted to ("news" by default); recommendations for
            every feed are tracked separately
        """
        location = "/recommendations/postings/accepted/{nodeName}/{postingId}".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id)
        )
        params = {"feed": feed}
        data = self.call(
            "accept_recommended_posting", location, method="POST", params=params, schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def reject_recommended_posting(self, node_name: str, posting_id: str, feed: str | None = None) -> types.Result:
        """
        Inform the recommendation service that the recommended posting was rejected by the client.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        :param feed: name of the feed the recommendation is rejected for ("news" by default); recommendations for
            every feed are tracked separately
        """
        location = "/recommendations/postings/rejected/{nodeName}/{postingId}".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id)
        )
        params = {"feed": feed}
        data = self.call(
            "reject_recommended_posting", location, method="POST", params=params, schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def get_recommended_nodes_by_activity(
        self, sheriff: str | None = None, limit: int | None = None
    ) -> List[types.RecommendedNodeInfo]:
        """
        Find the most active nodes known to the recommendation service.
        
        The service may decide to return fewer recommendations than the given ``limit``.

        :param sheriff: filter out nodes prohibited by the given sheriff
        :param limit: maximum number of recommendations returned
        """
        location = "/recommendations/nodes/active".format()
        params = {"sheriff": sheriff, "limit": limit}
        data = self.call(
            "get_recommended_nodes_by_activity", location, method="GET", auth=False, params=params,
            schema=schemas.RECOMMENDED_NODE_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.RecommendedNodeInfo)

    def exclude_node_from_recommendations(self, node_name: str) -> types.Result:
        """
        Ask the recommendation service to exclude all content from the given node from future recommendations.

        :param node_name: name of the remote node
        """
        location = "/recommendations/nodes/excluded/{nodeName}".format(nodeName=quote_plus(node_name))
        data = self.call("exclude_node_from_recommendations", location, method="POST", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def allow_node_in_recommendations(self, node_name: str) -> types.Result:
        """
        Allow the recommendation service to include the content from the given node to future recommendations.

        :param node_name: name of the remote node
        """
        location = "/recommendations/nodes/excluded/{nodeName}".format(nodeName=quote_plus(node_name))
        data = self.call("allow_node_in_recommendations", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def ask_remote_node(self, node_name: str, details: types.AskDescription) -> types.Result:
        """
        Send a request to the remote node.

        :param node_name: name of the remote node
        :param details:
        """
        location = "/nodes/{nodeName}/ask".format(nodeName=quote_plus(node_name))
        data = self.call("ask_remote_node", location, method="POST", body=details, schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def create_remote_comment(
        self, node_name: str, posting_id: str, comment: types.CommentSourceText
    ) -> types.Result:
        """
        Add a comment to the posting on the remote node and register it in the registry at the local node.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        :param comment:
        """
        location = "/nodes/{nodeName}/postings/{postingId}/comments".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id)
        )
        data = self.call(
            "create_remote_comment", location, method="POST", src_bodies=True, body=comment,
            schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def update_remote_comment(
        self, node_name: str, posting_id: str, comment_id: str, comment: types.CommentSourceText
    ) -> types.Result:
        """
        Update a comment to the posting on the remote node.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        :param comment_id: ID of the comment on the remote node
        :param comment:
        """
        location = "/nodes/{nodeName}/postings/{postingId}/comments/{commentId}".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        data = self.call(
            "update_remote_comment", location, method="PUT", src_bodies=True, body=comment,
            schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def delete_remote_comment(self, node_name: str, posting_id: str, comment_id: str) -> types.Result:
        """
        Delete a comment from the registry of all comments at the local node.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        :param comment_id: ID of the comment on the remote node
        """
        location = "/nodes/{nodeName}/postings/{postingId}/comments/{commentId}".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        data = self.call("delete_remote_comment", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def verify_remote_comment(self, node_name: str, posting_id: str, comment_id: str) -> types.AsyncOperationCreated:
        """
        Verify the signature of the given comment to the posting on the remote node.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        :param comment_id: ID of the comment on the remote node
        """
        location = "/nodes/{nodeName}/postings/{postingId}/comments/{commentId}/verify".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        data = self.call(
            "verify_remote_comment", location, method="POST", schema=schemas.ASYNC_OPERATION_CREATED_SCHEMA
        )
        return types.AsyncOperationCreated.from_json(data)

    def create_remote_comment_reaction(
        self, node_name: str, posting_id: str, comment_id: str, reaction: types.ReactionAttributes
    ) -> types.Result:
        """
        Add a reaction to the comment on the remote node and register it in the registry at the local node.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        :param comment_id: ID of the comment on the remote node
        :param reaction:
        """
        location = "/nodes/{nodeName}/postings/{postingId}/comments/{commentId}/reactions".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        data = self.call(
            "create_remote_comment_reaction", location, method="POST", body=reaction, schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def delete_remote_comment_reaction(self, node_name: str, posting_id: str, comment_id: str) -> types.Result:
        """
        Delete a reaction from the registry of all reactions at the local node.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        :param comment_id: ID of the comment on the remote node
        """
        location = "/nodes/{nodeName}/postings/{postingId}/comments/{commentId}/reactions".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id), commentId=quote_plus(comment_id)
        )
        data = self.call("delete_remote_comment_reaction", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def verify_remote_comment_reaction(
        self, node_name: str, posting_id: str, comment_id: str, owner_name: str
    ) -> types.AsyncOperationCreated:
        """
        Verify the signature of the reaction of the given owner to the comment on the remote node.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        :param comment_id: ID of the comment on the remote node
        :param owner_name: reaction owner node name
        """
        location = "/nodes/{nodeName}/postings/{postingId}/comments/{commentId}/reactions/{ownerName}/verify".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id), commentId=quote_plus(comment_id),
            ownerName=quote_plus(owner_name)
        )
        data = self.call(
            "verify_remote_comment_reaction", location, method="POST", schema=schemas.ASYNC_OPERATION_CREATED_SCHEMA
        )
        return types.AsyncOperationCreated.from_json(data)

    def create_remote_posting(self, node_name: str, posting: types.PostingSourceText) -> types.Result:
        """
        Add a posting to the remote node and register it in the registry at the local node.

        :param node_name: name of the remote node
        :param posting:
        """
        location = "/nodes/{nodeName}/postings".format(nodeName=quote_plus(node_name))
        data = self.call(
            "create_remote_posting", location, method="POST", src_bodies=True, body=posting,
            schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def update_remote_posting(
        self, node_name: str, posting_id: str, posting: types.PostingSourceText
    ) -> types.Result:
        """
        Update a posting on the remote node.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        :param posting:
        """
        location = "/nodes/{nodeName}/postings/{postingId}".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id)
        )
        data = self.call(
            "update_remote_posting", location, method="PUT", src_bodies=True, body=posting,
            schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def delete_remote_posting(self, node_name: str, posting_id: str) -> types.Result:
        """
        Delete a posting from the registry of all remote postings at the local node.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        """
        location = "/nodes/{nodeName}/postings/{postingId}".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id)
        )
        data = self.call("delete_remote_posting", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def verify_remote_posting(self, node_name: str, id: str) -> types.AsyncOperationCreated:
        """
        Verify the signature of the given posting.

        :param node_name: name of the remote node
        :param id: ID of the posting on the remote node
        """
        location = "/nodes/{nodeName}/postings/{id}/verify".format(nodeName=quote_plus(node_name), id=quote_plus(id))
        data = self.call(
            "verify_remote_posting", location, method="POST", schema=schemas.ASYNC_OPERATION_CREATED_SCHEMA
        )
        return types.AsyncOperationCreated.from_json(data)

    def verify_remote_posting_revision(
        self, node_name: str, id: str, revision_id: str
    ) -> types.AsyncOperationCreated:
        """
        Verify the signature of the given revision of a posting.

        :param node_name: name of the remote node
        :param id: ID of the posting on the remote node
        :param revision_id: ID of the posting revision
        """
        location = "/nodes/{nodeName}/postings/{id}/revisions/{revisionId}/verify".format(
            nodeName=quote_plus(node_name), id=quote_plus(id), revisionId=quote_plus(revision_id)
        )
        data = self.call(
            "verify_remote_posting_revision", location, method="POST", schema=schemas.ASYNC_OPERATION_CREATED_SCHEMA
        )
        return types.AsyncOperationCreated.from_json(data)

    def create_remote_posting_reaction(
        self, node_name: str, posting_id: str, reaction: types.ReactionAttributes
    ) -> types.Result:
        """
        Add a reaction to the posting on the remote node and register it in the registry at the local node.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        :param reaction:
        """
        location = "/nodes/{nodeName}/postings/{postingId}/reactions".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id)
        )
        data = self.call(
            "create_remote_posting_reaction", location, method="POST", body=reaction, schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def delete_remote_posting_reaction(self, node_name: str, posting_id: str) -> types.Result:
        """
        Delete a reaction from the registry of all reactions at the local node.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        """
        location = "/nodes/{nodeName}/postings/{postingId}/reactions".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id)
        )
        data = self.call("delete_remote_posting_reaction", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def verify_remote_posting_reaction(
        self, node_name: str, posting_id: str, owner_name: str
    ) -> types.AsyncOperationCreated:
        """
        Verify the signature of the reaction of the given owner to the posting on the remote node.

        :param node_name: name of the remote node
        :param posting_id: ID of the posting on the remote node
        :param owner_name: reaction owner node name
        """
        location = "/nodes/{nodeName}/postings/{postingId}/reactions/{ownerName}/verify".format(
            nodeName=quote_plus(node_name), postingId=quote_plus(posting_id), ownerName=quote_plus(owner_name)
        )
        data = self.call(
            "verify_remote_posting_reaction", location, method="POST", schema=schemas.ASYNC_OPERATION_CREATED_SCHEMA
        )
        return types.AsyncOperationCreated.from_json(data)

    def create_remote_sheriff_order(
        self, node_name: str, sheriff_order: types.SheriffOrderAttributes
    ) -> types.Result:
        """
        Sign and send the order to the remote node and store it in the registry at the local node.

        :param node_name: name of the remote node
        :param sheriff_order:
        """
        location = "/nodes/{nodeName}/sheriff/orders".format(nodeName=quote_plus(node_name))
        data = self.call(
            "create_remote_sheriff_order", location, method="POST", body=sheriff_order, schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def get_remote_sheriff_order(self, node_name: str, id: str) -> types.SheriffOrderInfo:
        """
        Get the details of the given sheriff's order

        :param node_name: name of the remote node
        :param id: ID of the order
        """
        location = "/nodes/{nodeName}/sheriff/orders/{id}".format(nodeName=quote_plus(node_name), id=quote_plus(id))
        data = self.call(
            "get_remote_sheriff_order", location, method="GET", auth=False, schema=schemas.SHERIFF_ORDER_INFO_SCHEMA
        )
        return types.SheriffOrderInfo.from_json(data)

    def search_nodes(self, filter: types.SearchNodeFilter) -> types.SearchNodePageInfo:
        """
        Search for Moera nodes matching the search ``query``. Every space-delimited word in the query must match
        case-insensitively a beginning of the node's name or a beginning of any non-letter-delimited word in the node's
        full name. The order of words is not significant.
        
        The search engine may decide to return fewer nodes than the given ``limit``.
        
        The returned nodes are sorted by their relevance. The exact definition of this term is left to the search
        engine's implementation.

        :param filter:
        """
        location = "/search/nodes"
        data = self.call(
            "search_nodes", location, method="POST", body=filter, schema=schemas.SEARCH_NODE_PAGE_INFO_SCHEMA
        )
        return types.SearchNodePageInfo.from_json(data)

    def search_node_suggestions(
        self, query: str | None = None, sheriff: str | None = None, limit: int | None = None
    ) -> List[types.SearchNodeInfo]:
        """
        Search for Moera nodes matching the search ``query`` and return a short list of "smart suggestions" for the
        user. Every space-delimited word in the query must match case-insensitively a beginning of the node's name or a
        beginning of any non-letter-delimited word in the node's full name. The order of words is not significant.
        
        The search engine may decide to return fewer nodes than the given ``limit``.
        
        The returned nodes are sorted by their relevance. The exact definition of this term is left to the search
        engine's implementation.

        :param query: the search query
        :param sheriff: filter out entries prohibited by the given sheriff
        :param limit: maximum number of nodes returned
        """
        location = "/search/nodes/suggestions".format()
        params = {"query": query, "sheriff": sheriff, "limit": limit}
        data = self.call(
            "search_node_suggestions", location, method="GET", params=params,
            schema=schemas.SEARCH_NODE_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.SearchNodeInfo)

    def search_entries_by_hashtag(self, filter: types.SearchHashtagFilter) -> types.SearchHashtagSliceInfo:
        """
        Search for postings and comments containing the specified hashtag(s) and optionally filtered by other criteria.
        
        The search engine may decide to return fewer nodes than the given ``limit``.
        
        The returned entries are sorted by moment in descending order.

        :param filter:
        """
        location = "/search/entries/by-hashtag"
        data = self.call(
            "search_entries_by_hashtag", location, method="POST", body=filter, bodies=True,
            schema=schemas.SEARCH_HASHTAG_SLICE_INFO_SCHEMA
        )
        return types.SearchHashtagSliceInfo.from_json(data)

    def search_entries_by_text(self, filter: types.SearchTextFilter) -> types.SearchTextPageInfo:
        """
        Search for postings and comments containing the specified words or text fragment, and optionally filtered by
        other criteria.
        
        The search engine may decide to return fewer nodes than the given ``limit``.
        
        The returned entries are sorted by their relevance. The exact definition of this term is left to the search
        engine's implementation.

        :param filter:
        """
        location = "/search/entries/by-text"
        data = self.call(
            "search_entries_by_text", location, method="POST", body=filter, bodies=True,
            schema=schemas.SEARCH_TEXT_PAGE_INFO_SCHEMA
        )
        return types.SearchTextPageInfo.from_json(data)

    def update_settings(self, settings: List[types.SettingInfo]) -> types.Result:
        """
        Update the given settings. If the input contains node settings, they are validated and the first validation
        error is returned, if any. The update is always performed as a whole - if there is an error saving any one of
        the settings in the input, none of them are updated.
        
        If one of the settings to be updated is privileged, *root secret* authentication is required. If one of the
        settings to be updated is non-privileged, *admin* authentication is required.

        :param settings:
        """
        location = "/settings"
        data = self.call("update_settings", location, method="PUT", body=settings, schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def get_client_settings(self, prefix: str | None = None) -> List[types.SettingInfo]:
        """
        Get all client settings, sorted by name.

        :param prefix: filter settings whose names start with the given prefix, case-sensitive (``client.`` prefix
            must be included)
        """
        location = "/settings/client".format()
        params = {"prefix": prefix}
        data = self.call(
            "get_client_settings", location, method="GET", params=params, schema=schemas.SETTING_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.SettingInfo)

    def get_node_settings(self, prefix: str | None = None) -> List[types.SettingInfo]:
        """
        Get all node settings, sorted by name. If a setting has not changed its value from the default, it is omitted.

        :param prefix: filter settings whose names start with the given prefix, case-sensitive
        """
        location = "/settings/node".format()
        params = {"prefix": prefix}
        data = self.call(
            "get_node_settings", location, method="GET", params=params, schema=schemas.SETTING_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.SettingInfo)

    def get_node_settings_metadata(self, prefix: str | None = None) -> List[types.SettingMetaInfo]:
        """
        Get all node settings metadata, sorted by name.

        :param prefix: filter settings whose names start with the given prefix, case-sensitive
        """
        location = "/settings/node/metadata".format()
        params = {"prefix": prefix}
        data = self.call(
            "get_node_settings_metadata", location, method="GET", params=params,
            schema=schemas.SETTING_META_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.SettingMetaInfo)

    def update_node_settings_metadata(self, metadata: List[types.SettingMetaAttributes]) -> types.Result:
        """
        Update node settings metadata, overriding built-in defaults.

        :param metadata:
        """
        location = "/settings/node/metadata"
        data = self.call(
            "update_node_settings_metadata", location, method="PUT", body=metadata, schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def get_sheriff_complaint_groups_slice(
        self, after: int | None = None, before: int | None = None, limit: int | None = None,
        status: types.SheriffComplaintStatus | None = None
    ) -> types.SheriffComplaintGroupsSliceInfo:
        """
        Get a slice of the list of groups of complaints, optionally filtered by status, delimited by the ``before`` or
        ``after`` moment and the given ``limit``. If neither ``before`` nor ``after`` are provided, the latest groups
        are returned. The node may decide to return fewer groups than the given ``limit``. The groups are always sorted
        by moment, descending.

        :param after: filter groups created strongly after this moment
        :param before: filter groups created at or before this moment
        :param limit: maximum number of groups returned
        :param status: filter groups by status
        """
        location = "/sheriff/complaints/groups".format()
        params = {"after": after, "before": before, "limit": limit, "status": status}
        data = self.call(
            "get_sheriff_complaint_groups_slice", location, method="GET", auth=False, params=params,
            schema=schemas.SHERIFF_COMPLAINT_GROUPS_SLICE_INFO_SCHEMA
        )
        return types.SheriffComplaintGroupsSliceInfo.from_json(data)

    def get_sheriff_complaint_group(self, id: str) -> types.SheriffComplaintGroupInfo:
        """
        Get details of the given group of complaints.

        :param id: ID of the group of complaints
        """
        location = "/sheriff/complaints/groups/{id}".format(id=quote_plus(id))
        data = self.call(
            "get_sheriff_complaint_group", location, method="GET", auth=False,
            schema=schemas.SHERIFF_COMPLAINT_GROUP_INFO_SCHEMA
        )
        return types.SheriffComplaintGroupInfo.from_json(data)

    def get_sheriff_complaints_by_group(self, id: str) -> List[types.SheriffComplaintInfo]:
        """
        Get complaints included in the given group of complaints.

        :param id: ID of the group of complaints
        """
        location = "/sheriff/complaints/groups/{id}/complaints".format(id=quote_plus(id))
        data = self.call(
            "get_sheriff_complaints_by_group", location, method="GET", auth=False,
            schema=schemas.SHERIFF_COMPLAINT_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.SheriffComplaintInfo)

    def update_sheriff_complaint_group(
        self, id: str, decision: types.SheriffComplaintDecisionText
    ) -> types.SheriffComplaintGroupInfo:
        """
        Make a decision on the given group of complaints.

        :param id: ID of the group of complaints
        :param decision:
        """
        location = "/sheriff/complaints/groups/{id}".format(id=quote_plus(id))
        data = self.call(
            "update_sheriff_complaint_group", location, method="PUT", body=decision,
            schema=schemas.SHERIFF_COMPLAINT_GROUP_INFO_SCHEMA
        )
        return types.SheriffComplaintGroupInfo.from_json(data)

    def create_sheriff_complaint(self, complaint: types.SheriffComplaintText) -> types.SheriffComplaintInfo:
        """
        Send a complaint to the sheriff.

        :param complaint:
        """
        location = "/sheriff/complaints"
        data = self.call(
            "create_sheriff_complaint", location, method="POST", body=complaint,
            schema=schemas.SHERIFF_COMPLAINT_INFO_SCHEMA
        )
        return types.SheriffComplaintInfo.from_json(data)

    def create_sheriff_order(self, sheriff_order: types.SheriffOrderDetails) -> types.Result:
        """
        Receive and execute the sheriff's order.

        :param sheriff_order:
        """
        location = "/sheriff/orders"
        data = self.call(
            "create_sheriff_order", location, method="POST", body=sheriff_order, auth=False,
            schema=schemas.RESULT_SCHEMA
        )
        return types.Result.from_json(data)

    def get_story(self, id: str) -> types.StoryInfo:
        """
        Get an individual story.

        :param id: ID of the story
        """
        location = "/stories/{id}".format(id=quote_plus(id))
        data = self.call("get_story", location, method="GET", bodies=True, schema=schemas.STORY_INFO_SCHEMA)
        return types.StoryInfo.from_json(data)

    def update_story(self, id: str, story: types.StoryAttributes) -> types.StoryInfo:
        """
        Update the story.

        :param id: ID of the story
        :param story:
        """
        location = "/stories/{id}".format(id=quote_plus(id))
        data = self.call(
            "update_story", location, method="PUT", body=story, bodies=True, schema=schemas.STORY_INFO_SCHEMA
        )
        return types.StoryInfo.from_json(data)

    def delete_story(self, id: str) -> types.Result:
        """
        Delete the story.

        :param id: ID of the story
        """
        location = "/stories/{id}".format(id=quote_plus(id))
        data = self.call("delete_story", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def get_subscribers(
        self, node_name: str | None = None, type: types.SubscriptionType | None = None, feed_name: str | None = None,
        entry_id: str | None = None
    ) -> List[types.SubscriberInfo]:
        """
        Get the list of all subscribers, optionally filtered by some criteria.

        :param node_name: filter by subscribed node name
        :param type: filter by subscription type
        :param feed_name: filter by name of the feed subscribed to
        :param entry_id: filter by ID of the entry subscribed to
        """
        location = "/people/subscribers".format()
        params = {"nodeName": node_name, "type": type, "feedName": feed_name, "entryId": entry_id}
        data = self.call(
            "get_subscribers", location, method="GET", params=params, schema=schemas.SUBSCRIBER_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.SubscriberInfo)

    def create_subscriber(self, subscriber: types.SubscriberDescription) -> types.SubscriberInfo:
        """
        Subscribe to a particular group of notifications.

        :param subscriber:
        """
        location = "/people/subscribers"
        data = self.call(
            "create_subscriber", location, method="POST", body=subscriber, schema=schemas.SUBSCRIBER_INFO_SCHEMA
        )
        return types.SubscriberInfo.from_json(data)

    def get_subscriber(self, id: str) -> types.SubscriberInfo:
        """
        Get an individual subscriber.

        :param id: ID of the subscriber
        """
        location = "/people/subscribers/{id}".format(id=quote_plus(id))
        data = self.call("get_subscriber", location, method="GET", schema=schemas.SUBSCRIBER_INFO_SCHEMA)
        return types.SubscriberInfo.from_json(data)

    def update_subscriber(self, id: str, subscriber: types.SubscriberOverride) -> types.SubscriberInfo:
        """
        Update the subscriber's operations or set operations' overrides.

        :param id: ID of the subscriber
        :param subscriber:
        """
        location = "/people/subscribers/{id}".format(id=quote_plus(id))
        data = self.call(
            "update_subscriber", location, method="PUT", body=subscriber, schema=schemas.SUBSCRIBER_INFO_SCHEMA
        )
        return types.SubscriberInfo.from_json(data)

    def delete_subscriber(self, id: str) -> types.ContactInfo:
        """
        Delete the subscriber and return the updated information about the node that was subscribed.

        :param id: ID of the subscriber
        """
        location = "/people/subscribers/{id}".format(id=quote_plus(id))
        data = self.call("delete_subscriber", location, method="DELETE", schema=schemas.CONTACT_INFO_SCHEMA)
        return types.ContactInfo.from_json(data)

    def get_subscriptions(
        self, node_name: str | None = None, type: types.SubscriptionType | None = None
    ) -> List[types.SubscriptionInfo]:
        """
        Get the list of all subscriptions, optionally filtered by some criteria.

        :param node_name: filter by node name
        :param type: filter by subscription type
        """
        location = "/people/subscriptions".format()
        params = {"nodeName": node_name, "type": type}
        data = self.call(
            "get_subscriptions", location, method="GET", params=params, schema=schemas.SUBSCRIPTION_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.SubscriptionInfo)

    def create_subscription(self, subscription: types.SubscriptionDescription) -> types.SubscriptionInfo:
        """
        Register a subscription to notifications from a particular node.

        :param subscription:
        """
        location = "/people/subscriptions"
        data = self.call(
            "create_subscription", location, method="POST", body=subscription, schema=schemas.SUBSCRIPTION_INFO_SCHEMA
        )
        return types.SubscriptionInfo.from_json(data)

    def update_subscription(self, id: str, subscription: types.SubscriptionOverride) -> types.SubscriptionInfo:
        """
        Update the subscription's operations or set operations' overrides.

        :param id: ID of the subscription
        :param subscription:
        """
        location = "/people/subscriptions/{id}".format(id=quote_plus(id))
        data = self.call(
            "update_subscription", location, method="PUT", body=subscription, schema=schemas.SUBSCRIPTION_INFO_SCHEMA
        )
        return types.SubscriptionInfo.from_json(data)

    def delete_subscription(self, id: str) -> types.ContactInfo:
        """
        Delete the subscription and return the updated information about the node that was subscribed to.

        :param id: ID of the subscription
        """
        location = "/people/subscriptions/{id}".format(id=quote_plus(id))
        data = self.call("delete_subscription", location, method="DELETE", schema=schemas.CONTACT_INFO_SCHEMA)
        return types.ContactInfo.from_json(data)

    def search_subscriptions(self, filter: types.SubscriptionFilter) -> List[types.SubscriptionInfo]:
        """
        Search for subscriptions by the given criteria.

        :param filter:
        """
        location = "/people/subscriptions/search"
        data = self.call(
            "search_subscriptions", location, method="POST", body=filter, schema=schemas.SUBSCRIPTION_INFO_ARRAY_SCHEMA
        )
        return structure_list(data, types.SubscriptionInfo)

    def get_tokens(self) -> List[types.TokenInfo]:
        """
        Get the list of all existing tokens.
        """
        location = "/tokens"
        data = self.call("get_tokens", location, method="GET", schema=schemas.TOKEN_INFO_ARRAY_SCHEMA)
        return structure_list(data, types.TokenInfo)

    def create_token(self, token: types.TokenAttributes) -> types.TokenInfo:
        """
        Sign in and create a token.

        :param token:
        """
        location = "/tokens"
        data = self.call(
            "create_token", location, method="POST", body=token, auth=False, schema=schemas.TOKEN_INFO_SCHEMA
        )
        return types.TokenInfo.from_json(data)

    def get_token_info(self, id: str) -> types.TokenInfo:
        """
        Get information about the token.

        :param id: ID of the token
        """
        location = "/tokens/{id}".format(id=quote_plus(id))
        data = self.call("get_token_info", location, method="GET", schema=schemas.TOKEN_INFO_SCHEMA)
        return types.TokenInfo.from_json(data)

    def update_token(self, id: str, update: types.TokenUpdate) -> types.TokenInfo:
        """
        Update the name or permissions of the token. It is not possible to grant token additional permissions with this
        request, but the granted permissions can be revoked.

        :param id: ID of the token
        :param update:
        """
        location = "/tokens/{id}".format(id=quote_plus(id))
        data = self.call("update_token", location, method="PUT", body=update, schema=schemas.TOKEN_INFO_SCHEMA)
        return types.TokenInfo.from_json(data)

    def delete_token(self, id: str) -> types.Result:
        """
        Delete the token.

        :param id: ID of the token
        """
        location = "/tokens/{id}".format(id=quote_plus(id))
        data = self.call("delete_token", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def get_user_list_general(self, name: str) -> types.UserListInfo:
        """
        Get the general information about the user list given.

        :param name: the name of the list
        """
        location = "/user-lists/{name}".format(name=quote_plus(name))
        data = self.call(
            "get_user_list_general", location, method="GET", auth=False, schema=schemas.USER_LIST_INFO_SCHEMA
        )
        return types.UserListInfo.from_json(data)

    def get_user_list_slice(
        self, name: str, after: int | None = None, before: int | None = None, limit: int | None = None
    ) -> types.UserListSliceInfo:
        """
        Get a slice of the user list, delimited by the ``before`` or ``after`` moment and the given ``limit``. If
        neither ``before`` nor ``after`` are provided, the latest items are returned. The node may decide to return
        fewer items than the given ``limit``. The items are always sorted by moment, descending.

        :param name: the name of the list
        :param after: filter items created strongly after this moment
        :param before: filter items created at or before this moment
        :param limit: maximum number of items returned
        """
        location = "/user-lists/{name}/items".format(name=quote_plus(name))
        params = {"after": after, "before": before, "limit": limit}
        data = self.call(
            "get_user_list_slice", location, method="GET", auth=False, params=params,
            schema=schemas.USER_LIST_SLICE_INFO_SCHEMA
        )
        return types.UserListSliceInfo.from_json(data)

    def get_user_list_item(self, name: str, node_name: str) -> types.UserListItemInfo:
        """
        Get the information from the user list about the node given.

        :param name: the name of the list
        :param node_name: the node name to get information about
        """
        location = "/user-lists/{name}/items/{nodeName}".format(name=quote_plus(name), nodeName=quote_plus(node_name))
        data = self.call(
            "get_user_list_item", location, method="GET", auth=False, schema=schemas.USER_LIST_ITEM_INFO_SCHEMA
        )
        return types.UserListItemInfo.from_json(data)

    def create_user_list_item(self, name: str, item: types.UserListItemAttributes) -> types.UserListItemInfo:
        """
        Add a node to the user list.

        :param name: the name of the list
        :param item:
        """
        location = "/user-lists/{name}/items".format(name=quote_plus(name))
        data = self.call(
            "create_user_list_item", location, method="POST", body=item, schema=schemas.USER_LIST_ITEM_INFO_SCHEMA
        )
        return types.UserListItemInfo.from_json(data)

    def delete_user_list_item(self, name: str, node_name: str) -> types.Result:
        """
        Delete a node from the user list

        :param name: the name of the list
        :param node_name: the node name to delete
        """
        location = "/user-lists/{name}/items/{nodeName}".format(name=quote_plus(name), nodeName=quote_plus(node_name))
        data = self.call("delete_user_list_item", location, method="DELETE", schema=schemas.RESULT_SCHEMA)
        return types.Result.from_json(data)

    def who_am_i(self) -> types.WhoAmI:
        """
        Get brief information about the node.
        """
        location = "/whoami"
        data = self.call("who_am_i", location, method="GET", auth=False, schema=schemas.WHO_AM_I_SCHEMA)
        return types.WhoAmI.from_json(data)



from __future__ import annotations

import re
import sys
from typing import Any, TextIO, Mapping, List, Sequence

import yaml
from camel_converter import to_snake


def ind(n: int) -> str:
    return n * 4 * ' '


def doc_wrap(s: str, indent: int) -> str:
    s = s.strip()
    if '\n' in s:
        return f'\n{ind(indent)}'.join(doc_wrap(c, indent) for c in s.split('\n'))

    max_length = 120 - indent * 4
    result = ''
    while True:
        if len(s) < max_length:
            result += s
            break
        pos = 0
        while True:
            next = s.find(' ', pos + 1)
            if next < 0 or next >= max_length:
                break
            pos = next
        result += s[:pos] + '\n' + ind(indent)
        s = s[pos + 1:]
    return result


def comma_wrap(s: str, indent: int) -> str:
    max_length = 120 - indent * 4
    result = ''
    while True:
        if len(s) < max_length:
            result += s
            break
        pos = 0
        while True:
            next = s.find(', ', pos + 1)
            if next < 0 or next >= max_length:
                break
            pos = next
        result += s[:pos] + ',\n' + ind(indent)
        s = s[pos + 2:]
    return result


def params_wrap(template: str, substitute: str, indent: int) -> str:
    line = template % substitute
    if len(line) > 120:
        line = template % ('\n' + ind(indent) + comma_wrap(substitute, indent) + '\n' + ind(indent - 1))
    return line


def html_to_doc(s: str) -> str:
    s = (s.replace('<code>', '``')
          .replace('</code>', '``')
          .replace('<br>', '\n')
          .replace('<i>', '*')
          .replace('</i>', '*')
          .replace('<b>', '**')
          .replace('</b>', '**')
          .replace('<ul>', '\n')
          .replace('<li>', '* ')
          .replace('</li>', '')
          .replace('</ul>', '')
          .replace('``null``', '``None``')
          .replace('``true``', '``True``')
          .replace('``false``', '``False``'))
    s = re.sub(r'<a [^>]*>', '', s)
    s = s.replace('</a>', '')
    return s


def read_api(ifname: str) -> Any:
    with open(ifname, 'r') as ifile:
        return yaml.safe_load(ifile)


def generate_enum(enum: Any, tfile: TextIO) -> None:
    values = ", ".join(f'"{item["name"]}"' for item in enum['values'])
    line = params_wrap(f'\n{enum["name"]} = Literal[%s]\n', values, 1)
    tfile.write(line)
    if any('value' in item for item in enum['values']):
        tfile.write(f'\n{to_snake(enum["name"]).upper()}_VALUES: Mapping[{enum["name"]}, int] = {{\n')
        for item in enum['values']:
            tfile.write(f'{ind(1)}"{item["name"]}": {item["value"]},\n')
        tfile.write('}\n')


def schema_type(sfile: TextIO, indent: int, a_type: str, struct: bool = False, nullable: bool = False,
                default: Any = None, min: float | None = None, max: float | None = None) -> None:
    if struct:
        if nullable:
            sfile.write('to_nullable_object_schema(%s_SCHEMA)' % to_snake(a_type).upper())
        else:
            sfile.write(to_snake(a_type).upper() + '_SCHEMA')
        return

    sfile.write('{\n')
    if nullable:
        sfile.write(ind(indent + 1) + f'"type": ["{a_type}", "null"]')
    else:
        sfile.write(ind(indent + 1) + f'"type": "{a_type}"')
    if default is not None:
        sfile.write(',\n')
        sfile.write(ind(indent + 1) + f'"default": {default}')
    if min is not None:
        sfile.write(',\n')
        sfile.write(ind(indent + 1) + f'"minimum": {min}')
    if max is not None:
        sfile.write(',\n')
        sfile.write(ind(indent + 1) + f'"maximum": {max}')
    sfile.write('\n')
    sfile.write(ind(indent) + '}')


def schema_array(sfile: TextIO, indent: int, a_type: str, struct: bool = False, nullable: bool = False,
                 default: Any = None, min_items: int | None = None, max_items: int | None = None,
                 min: float | None = None, max: float | None = None) -> None:
    sfile.write('{\n')
    if nullable:
        sfile.write(ind(indent + 1) + '"type": ["array", "null"],\n')
    else:
        sfile.write(ind(indent + 1) + '"type": "array",\n')
    sfile.write(ind(indent + 1) + '"items": ')
    schema_type(sfile, indent + 1, a_type, struct=struct, nullable=False, min=min, max=max)
    if default is not None:
        sfile.write(',\n')
        sfile.write(ind(indent + 1) + f'"default": {default}')
    if min_items is not None:
        sfile.write(',\n')
        sfile.write(ind(indent + 1) + f'"minItems": {min_items}')
    if max_items is not None:
        sfile.write(',\n')
        sfile.write(ind(indent + 1) + f'"maxItems": {max_items}')
    sfile.write('\n')
    sfile.write(ind(indent) + '}')


def schema_map_string_int(sfile: TextIO, indent: int, nullable: bool = False) -> None:
    sfile.write('{\n')
    sfile.write(ind(indent + 1) + '"type": ' + ('["object", "null"]' if nullable else 'object') + ',\n')
    sfile.write(ind(indent + 1) + '"patternProperties": {\n')
    sfile.write(ind(indent + 2) + '"^.*$": ')
    schema_type(sfile, indent + 2, 'integer')
    sfile.write('\n')
    sfile.write(ind(indent + 1) + '}')
    sfile.write('\n')
    sfile.write(ind(indent) + '}')


def generate_operations(operations: Any, tfile: TextIO, sfile: TextIO) -> None:
    tfile.write(f'\n\nclass {operations["name"]}(Structure):\n')
    for field in operations['fields']:
        py_name = to_snake(field['name'])
        tfile.write('    %s: PrincipalValue | None = None\n' % py_name)
        if 'description' in field:
            tfile.write('    """%s"""\n' % field['description'])
        if 'default' in field:
            tfile.write('\n    @property\n')
            tfile.write('    def %s_or_default(self) -> PrincipalValue:\n' % py_name)
            doc = doc_wrap(
                field['description'] + ' (this property always returns default value instead of ``None``)', 2)
            if len(doc) >= 107:
                doc = f'\n{ind(2)}{doc}\n{ind(2)}'
            tfile.write('        """%s"""\n' % doc)
            tfile.write('        return self.%s if self.%s is not None else "%s"\n'
                        % (py_name, py_name, field['default']))

    sfile.write('\n')
    sfile.write('{name}_SCHEMA: Any = {{\n'.format(name=to_snake(operations['name']).upper()))
    sfile.write('    "type": "object",\n')
    sfile.write('    "properties": {\n')
    for field in operations['fields']:
        sfile.write(f'{ind(2)}"{field["name"]}": ')
        schema_type(sfile, 2, "string", nullable=True)
        sfile.write(',\n')
    sfile.write('    },\n')
    sfile.write('    "additionalProperties": False\n')
    sfile.write('}\n')


PY_TYPES = {
    'String': 'str',
    'String[]': 'List[str]',
    'short': 'int',
    'int': 'int',
    'long': 'int',
    'float': 'float',
    'boolean': 'bool',
    'timestamp': 'Timestamp',
    'byte[]': 'bytes',
    'UUID': 'str',
    'String -> int': 'Mapping[str, int]'
}


def to_py_type(api_type: str) -> str:
    py_type = PY_TYPES.get(api_type)
    if py_type is None:
        print('Unrecognized field type: ' + api_type)
        exit(1)
    return py_type


SCHEMA_TYPES = {
    'String': ('string', False),
    'String[]': ('string', True),
    'short': ('integer', False),
    'int': ('integer', False),
    'long': ('integer', False),
    'float': ('number', False),
    'boolean': ('boolean', False),
    'timestamp': ('integer', False),
    'byte[]': ('string', False),
    'UUID': ('string', False),
    'String -> int': schema_map_string_int
}


class Structure:
    data: Any
    generated: bool = False
    depends: Sequence[str]
    uses_body: bool = False
    output: bool = False
    output_array: bool = False

    def __init__(self, data: Any):
        self.data = data
        self.depends = [field['struct'] for field in data['fields'] if 'struct' in field]

    @property
    def generic(self) -> bool:
        return self.uses_body and self.output

    def generate_class(self, tfile: TextIO) -> None:
        tfile.write(f'\n\nclass {self.data["name"]}(Structure):\n')
        for field in self.data['fields']:
            if field.get('optional', False) and 'py-default' not in field:
                tmpl = '    %s: %s | None = None\n'
            else:
                tmpl = '    %s: %s\n'
            if 'struct' in field:
                t = field['struct']
            elif 'enum' in field:
                t = field['enum']
            else:
                if field['type'] == 'any':
                    continue
                t = to_py_type(field['type'])
            if field.get('array', False):
                t = f'List[{t}]'
            tfile.write(tmpl % (to_snake(field['name']), t))
            if 'description' in field:
                doc = doc_wrap(html_to_doc(field["description"]), 1)
                if len(doc) >= 107:
                    doc = f'\n{ind(1)}{doc}\n{ind(1)}'
                tfile.write(f'{ind(1)}"""{doc}"""\n')

    def generate_schema(self, sfile: TextIO) -> None:
        sfile.write('\n{name}_SCHEMA: Any = {{\n'
                    .format(name=to_snake(self.data['name']).upper()))
        sfile.write('    "type": "object",\n')
        sfile.write('    "properties": {\n')
        required: List[str] = []
        for field in self.data['fields']:
            if field.get('type') == 'any':
                continue

            sfile.write(f'{ind(2)}"{field["name"]}": ')
            default = field.get('py-default')
            optional = field.get('optional', False) and default is None
            array = field.get('array', False)
            if not optional:
                required.append(field['name'])
            struct = False
            if 'struct' in field:
                if field['struct'] == 'Body':
                    t = 'string'
                else:
                    t = field['struct']
                    struct = True
            elif 'enum' in field:
                t = 'string'
            else:
                s_type = SCHEMA_TYPES.get(field['type'])
                if callable(s_type):
                    t = None
                    s_type(sfile, 2, nullable=optional)
                else:
                    assert isinstance(s_type, tuple)
                    t, array = s_type
            constraints = field_constraints(field)
            if t is not None:
                if array:
                    schema_array(sfile, 2, t, struct=struct, nullable=optional, default=default,
                                 min_items=constraints.get('min-items'), max_items=constraints.get('max-items'),
                                 min=constraints.get('min'), max=constraints.get('max'))
                else:
                    schema_type(sfile, 2, t, struct=struct, nullable=optional, default=default,
                                min=constraints.get('min'), max=constraints.get('max'))
            sfile.write(',\n')
        sfile.write('    },\n')
        if len(required) > 0:
            sfile.write('    "required": [\n')
            for name in required:
                sfile.write(f'        "{name}",\n')
            sfile.write('    ],\n')
        sfile.write('    "additionalProperties": False\n')
        sfile.write('}\n')

        if self.output_array:
            sfile.write('\n{name}_ARRAY_SCHEMA = array_schema({name}_SCHEMA)\n'
                        .format(name=to_snake(self.data['name']).upper()))

    def generate(self, tfile: TextIO, sfile: TextIO) -> None:
        self.generate_class(tfile)
        if self.output:
            self.generate_schema(sfile)
        self.generated = True


def field_constraints(field: Any) -> dict:
    if 'constraints' not in field:
        return {}

    result = {}
    for constraint in field['constraints']:
        if 'value' in constraint:
            cons = constraint['value']
            if 'min' in cons:
                result['min'] = cons['min']
            if 'max' in cons:
                result['max'] = cons['max']
        if 'items' in constraint:
            cons = constraint['items']
            if 'min' in cons:
                result['min-items'] = cons['min']
            if 'max' in cons:
                result['max-items'] = cons['max']
    return result


def scan_body_usage(structs: Mapping[str, Structure]) -> None:
    for struct in structs.values():
        if 'Body' in struct.depends:
            struct.uses_body = True

    modified = True
    while modified:
        modified = False
        for struct in structs.values():
            if struct.uses_body:
                continue
            for dep in struct.depends:
                if dep in structs and structs[dep].uses_body:
                    struct.uses_body = True
                    modified = True


def scan_output_usage(api: Any, structs: Mapping[str, Structure]) -> None:
    for obj in api['objects']:
        for request in obj.get('requests', []):
            if 'out' not in request:
                continue
            if 'struct' not in request['out']:
                continue
            struct = request['out']['struct']
            if struct not in structs:
                continue
            structs[struct].output = True
            structs[struct].output_array |= request['out'].get('array', False)

    modified = True
    while modified:
        modified = False
        for struct in structs.values():
            if not struct.output:
                continue
            for dep in struct.depends:
                if dep in structs and not structs[dep].output:
                    structs[dep].output = True
                    modified = True


def scan_structures(api: Any) -> Mapping[str, Structure]:
    structs: Mapping[str, Structure] = {struct['name']: Structure(struct) for struct in api['structures']}
    scan_body_usage(structs)
    scan_output_usage(api, structs)
    return structs


def generate_structures(structs: Mapping[str, Structure], tfile: TextIO, sfile: TextIO) -> None:
    gen = True
    while gen:
        gen = False
        for struct in structs.values():
            if struct.generated:
                continue
            if any(not structs[d].generated for d in struct.depends if d in structs):
                continue
            struct.generate(tfile, sfile)
            gen = True
    loop = [s.data['name'] for s in structs.values() if not s.generated]
    if len(loop) > 0:
        print('Dependency loop in structures: ' + ', '.join(loop))
        exit(1)


def is_no_auth(auth: str) -> bool:
    variants = auth.split(" or ")
    return variants == ['none'] or variants == ['signature']


def generate_calls(api: Any, structs: Mapping[str, Structure], afile: TextIO) -> None:
    for obj in api['objects']:
        for request in obj.get('requests', []):
            if 'function' not in request:
                continue

            function = to_snake(request['function'])
            params = 'self'
            param_docs = []
            call_params = f'"{function}", location, method="{request["type"]}"'

            tail_params = ''
            url_params: dict[str, str] = {}
            flag_name: str | None = None
            flag_py_name: str | None = None
            flags: Sequence[str] = []
            for param in request.get('params', []) + request.get('query', []):
                if 'name' not in param:
                    print('Missing name of parameter of the request "{method} {url}"'
                          .format(method=request['type'], url=request['url']))
                    exit(1)
                name = param['name']
                py_name = to_snake(name)
                url_params[name] = py_name
                if 'enum' in param:
                    py_type = 'types.' + param['enum']
                else:
                    py_type = to_py_type(param['type'])
                if 'flags' in param:
                    flag_name = name
                    flag_py_name = py_name
                    flags = [flag['name'] for flag in param['flags']]
                    for flag in param['flags']:
                        flag_param = f'with_{flag["name"]}'
                        tail_params += f', {flag_param}: bool = False'
                        param_docs += [(flag_param, 'include ' + flag.get('description', ''))]
                else:
                    if param.get('optional', False):
                        tail_params += f', {py_name}: {py_type} | None = None'
                    else:
                        params += f', {py_name}: {py_type}'
                    param_docs += [(py_name, param.get('description', ''))]
            if 'in' in request:
                inp = request['in']
                if 'type' in inp:
                    if inp['type'] != 'blob':
                        print('Unrecognised type "{type}" of the input body of the request "{method} {url}"'
                              .format(type=inp['type'], method=request['type'], url=request['url']))
                        exit(1)
                    params += ', file: IO, file_type: str'
                    param_docs += [('file', ''), ('file_type', 'content-type of ``file``')]
                    call_params += f', body_file=file, body_file_type=file_type'
                else:
                    if 'name' not in inp:
                        print('Missing name of body of the request "{method} {url}"'
                              .format(method=request['type'], url=request['url']))
                        exit(1)
                    name = to_snake(inp['name'])
                    struct = inp['struct']
                    py_type = 'types.' + struct
                    if struct in structs and structs[struct].uses_body:
                        call_params += f', src_bodies=True'
                    if inp.get('array', False):
                        py_type = f'List[{py_type}]'
                    params += f', {name}: {py_type}'
                    param_docs += [(name, '')]
                    call_params += f', body={name}'
            params += tail_params

            if is_no_auth(request.get('auth', 'none')):
                call_params += ', auth=False'

            description = ''
            if 'description' in request:
                description = doc_wrap(html_to_doc(request["description"]), 2)
            if len(param_docs) > 0:
                description += '\n'
                for pd in param_docs:
                    doc = doc_wrap(f':param {pd[0]}: {html_to_doc(pd[1])}', 3)
                    description += f'\n{ind(2)}{doc}'
            if description != '':
                description = f'{ind(2)}"""\n{ind(2)}{description}\n{ind(2)}"""\n'

            location: str = request['url']
            if len(url_params) > 0:
                uparams = []
                p = re.compile(r'{(\w+)}')
                for name in p.findall(location):
                    if name not in url_params:
                        print('Unknown parameter "{param}" referenced in location "{url}"'
                              .format(param=name, url=request['url']))
                        exit(1)
                    uparams.append(f'{name}=quote_plus({url_params[name]})')
                    del url_params[name]
                location = params_wrap(
                    f'{ind(2)}location = "{location}".format(%s)\n',
                    ', '.join(uparams),
                    3
                )
            else:
                location = f'{ind(2)}location = "{location}"\n'

            subs = []
            for name, py_name in url_params.items():
                subs.append(f'"{name}": {py_name}')
            if len(subs) > 0:
                query_params = params_wrap(f'{ind(2)}params = {{%s}}\n', ', '.join(subs), 3)
                call_params += ', params=params'
            else:
                query_params = ''

            result = 'types.Result'
            result_schema = 'schemas.RESULT_SCHEMA'
            result_array = False
            if 'out' in request:
                out = request['out']
                if 'type' in out:
                    if out['type'] != 'blob':
                        print('Unrecognised type "{type}" of the output body of the request "{method} {url}"'
                              .format(type=out['type'], method=request['type'], url=request['url']))
                        exit(1)
                    result = 'IO'
                    result_schema = '"blob"'
                else:
                    struct = out['struct']
                    result = 'types.' + struct
                    if struct in structs and structs[struct].uses_body:
                        call_params += ', bodies=True'
                    if out.get('array', False):
                        result_schema = 'schemas.%s_ARRAY_SCHEMA' % to_snake(struct).upper()
                        result_array = True
                    else:
                        result_schema = 'schemas.%s_SCHEMA' % to_snake(struct).upper()
            call_params += f', schema={result_schema}'

            if result_array:
                afile.write(params_wrap(f'\n{ind(1)}def {function}(%s) -> List[{result}]:\n', params, 2))
            else:
                afile.write(params_wrap(f'\n{ind(1)}def {function}(%s) -> {result}:\n', params, 2))
            afile.write(description)
            afile.write(location)
            if flag_name is not None:
                items = ', '.join(f'"{flag}": with_{flag}' for flag in flags)
                afile.write(f'{ind(2)}{flag_py_name} = comma_separated_flags({{{items}}})\n')
            afile.write(query_params)
            afile.write(params_wrap(f'{ind(2)}data = self.call(%s)\n', call_params, 3))
            if result == 'IO':
                afile.write(f"{ind(2)}return cast(IO, data)\n")
            elif result_array:
                afile.write(f"{ind(2)}return structure_list(data, {result})\n")
            else:
                afile.write(f"{ind(2)}return {result}.from_json(data)\n")


PREAMBLE_TYPES = '''# This file is generated

from typing import List, Literal, Mapping, TypeAlias

from ..structure import Structure

Timestamp: TypeAlias = int
PrincipalValue: TypeAlias = str
'''

PREAMBLE_SCHEMAS = '''# This file is generated

from typing import Any

from ..structure import to_nullable_object_schema, array_schema
'''

PREAMBLE_CALLS = '''# This file is generated

from typing import IO, List, cast
from urllib.parse import quote_plus

from . import types, schemas
from .caller import Caller
from ..structure import comma_separated_flags, structure_list


class MoeraNode(Caller):
    """Node API interface."""

    def __init__(self, node_url: str | None = None):
        if node_url is not None:
            self.node_url(node_url)
'''


def generate_types(api: Any, outdir: str) -> None:
    structs = scan_structures(api)

    with open(outdir + '/types.py', 'w+') as tfile:
        with open(outdir + '/schemas.py', 'w+') as sfile:
            tfile.write(PREAMBLE_TYPES)
            sfile.write(PREAMBLE_SCHEMAS)
            for enum in api['enums']:
                generate_enum(enum, tfile)
            for operations in api['operations']:
                generate_operations(operations, tfile, sfile)
            generate_structures(structs, tfile, sfile)

    with open(outdir + '/node.py', 'w+') as afile:
        afile.write(PREAMBLE_CALLS)
        generate_calls(api, structs, afile)


FP_TYPES = {
    'String': 'string',
    'InetAddress': 'string',
    'int': 'number',
    'long': 'number',
    'timestamp': 'number',
    'byte': 'number',
    'byte[]': 'bytes',
    'boolean': 'boolean'
}


def generate_fingerprint_schema(schema: Any, name: str, version: int, ffile: TextIO) -> None:
    ffile.write(f'\n\n{to_snake(name).upper()}_FINGERPRINT{version}_SCHEMA: FingerprintSchema = [\n')
    ffile.write(f"{ind(1)}('version', 'number'),\n")
    for field in schema['fingerprint']:
        if 'type' in field:
            field_type = FP_TYPES[field['type']]
        else:
            field_type = 'bytes'
        if field.get('array', False):
            field_type += '[]'
        field_name = to_snake(field['field'])
        ffile.write(f"{ind(1)}('{field_name}', '{field_type}'),\n")
    ffile.write(']\n\n\n')


PY_FP_TYPES = {
    'String': 'str',
    'InetAddress': 'str',
    'int': 'int',
    'long': 'int',
    'timestamp': 'Timestamp',
    'byte': 'int',
    'byte[]': 'bytes',
    'boolean': 'bool'
}


def generate_fingerprint_function(schema: Any, name: str, version: int, ffile: TextIO) -> None:
    params = []
    object_type = ''
    for field in schema['fingerprint']:
        field_name = to_snake(field['field'])
        if field_name != 'object_type':
            if 'type' in field:
                field_type = PY_FP_TYPES[field['type']]
            else:
                field_type = 'bytes'
            if field.get('array', False):
                field_type = f'List[{field_type}]'
            params.append(f'{field_name}: {field_type} | None')
        else:
            value = field['comment'].removeprefix('<code>').removesuffix('</code>')
            object_type = f", 'object_type': '{value}'"
    ffile.write(params_wrap(f'def create_{to_snake(name)}_fingerprint{version}(%s) -> bytes:\n', ', '.join(params), 1))
    schema_name = to_snake(name).upper()
    call_params = f"{{'version': {version}{object_type}}} | locals(), {schema_name}_FINGERPRINT{version}_SCHEMA"
    ffile.write(params_wrap(f'{ind(1)}return fingerprint_bytes(%s)\n', call_params, 2))


def generate_fingerprint(schema: Any, name: str, version: int, ffile: TextIO) -> None:
    generate_fingerprint_schema(schema, name, version, ffile)
    generate_fingerprint_function(schema, name, version, ffile)


PREAMBLE_FINGERPRINTS = '''# This file is generated

from typing import List

from ..crypto.crypto import fingerprint_bytes
from ..crypto.fingerprint import FingerprintSchema
from .types import Timestamp'''


def generate_fingerprints(fp: Any, outdir: str) -> None:
    with open(outdir + '/fingerprints.py', 'w+') as ffile:
        ffile.write(PREAMBLE_FINGERPRINTS)
        for object in fp['objects']:
            name = object['name']
            for schema in object['versions']:
                version = schema['version']
                generate_fingerprint(schema, name, version, ffile)


if len(sys.argv) < 3 or sys.argv[1] == '':
    print("Usage: py-moera-api <node_api.yml file path> <node_api_fingerprints.yml file path> <output directory>")
    exit(1)

api = read_api(sys.argv[1])
fp = read_api(sys.argv[2])
outdir = sys.argv[3] if len(sys.argv) >= 4 else '.'
generate_types(api, outdir)
generate_fingerprints(fp, outdir)



from .node import MoeraNode
from .caller import NodeAuth, MoeraNodeError, MoeraNodeApiError, MoeraNodeConnectionError, moera_root
from .cartes import MoeraCarteSource, MoeraCartesError, generate_carte



from setuptools import setup

setup(
    name='moeralib',
    version='0.17.11',
    install_requires=[
        "requests>=2.32.0",
        'camel-converter',
        'jsonschema>=4.19.0',
        "cryptography~=43.0.1",
        'mnemonic',
        'ecpy',
    ],
)
